""" Constants for use in kerberos key generation and ingestion """
from ms_active_directory.environment.security.security_config_constants import ADEncryptionType

# constants needed to ingest keytabs and also write them out to files
# See these pages for more info on the evolution of the format beyond code comments
# http://manpages.ubuntu.com/manpages/artful/man3/krb5_fileformats.3.html
# https://web.mit.edu/kerberos/www/krb5-latest/doc/formats/keytab_file_format.html
ENCRYPTION_TYPE_FIELD_SIZE = 2  # uint16_t encryption_type = 16 bits to represent enc type = 2 bytes
ENTRY_LENGTH_FIELD_SIZE_BYTES = 4  # int32_t size of entry = 32 bits in entry size = 4 bytes
FLAGS_FIELD_SIZE_BYTES = 4  # uint32_t flags = 32 bits = 4 bytes
KEYTAB_STANDARD_LEADING_BYTES_SIZE = 1  # 0x5, a single byte, prefixes all keytab files for Kerberos v5
KEYTAB_FORMAT_SIZE_BYTES = 1  # uint8_t = 8 bits = 1 byte
KEY_LENGTH_FIELD_SIZE_BYTES = 2  # uint16_t key_length = 16 bits = 2 bytes
NUM_COMPONENTS_FIELD_SIZE_BYTES = 2  # uint16_t num_components = 16 bits = 2 bytes
REALM_LENGTH_FIELD_SIZE_BYTES = 2  # uint16_t = 16 bits in realm length = 2 bytes
PRINCIPAL_COMPONENT_LENGTH_FIELD_SIZE_BYTES = 2  # uint16_t = 16 bits in each component length = 2 bytes
PRINCIPAL_TYPE_FIELD_SIZE_BYTES = 4  # uint32_t name_type = 32 bits in name_type = 4 bytes
TIMESTAMP_FIELD_SIZE_BYTES = 4  # timestamp (time key was written) = 32 bits = 4 bytes
VNO8_FIELD_SIZE_BYTES = 1  # uint8_t vno8 = 8 bits in kvno = 1 byte
VNO32_FIELD_SIZE_BYTES = 4  # uint32_t vno32 = 32 bits in kvno = 1 byte

# this is a constant so that people can discover this comment explaining this weirdness.
# while an entire keytab file is encoded using some keytab format version, that keytab format version obviously cannot
# be or you wouldn't know how to read it.
# keytab format version is only 1 or 2, differentiating big and little endian encodings. when keytab format version
# itself is encoded, it's always encoded using format version 1 for backwards compatibility and discoverability
# purposes.
# we can be strongly certain that this will never change in the lifetime of kerberos version 5 as it breaks the ability
# of parsers to discover how to read the keytab itself
KEYTAB_FORMAT_VERSION_FOR_KEYTAB_FORMAT_VERSION = 1
# kerberos version 5 keytabs always start with 5
LEADING_BYTE_FOR_KERBEROS_V5 = 0x05
LEADING_HEX_FOR_KERBEROS_V5 = '05'
# format version 2 is newer and is preferred as it encodes more information and writes numbers in network byte order
PREFERRED_KEYTAB_FORMAT_VERSION = 2
# principal components are separated by forwarded slashes (not encoded)
PRINCIPAL_COMPONENT_DIVIDER = '/'


# here begin constants used to generate raw kerberos keys
AES_CIPHER_BLOCK_SIZE_BYTES = 16
AES_ITERATIONS_FOR_AD = 4096

SALT_FORMAT_FOR_AD = '{uppercase_realm}host{lowercase_computer_name}.{lowercase_domain}'

# AD uses a bitstring to encode supported encryption types, so the values for
# encryption types in AD are not the same as the values for encryption types
# in a kerberos keytab.
# this dictionary, and the ones below, help to make the translations of
# AD bitstring -> AD Enc Type -> KRB5 Enc Type Name -> KRB5 Enc Type Value
# and the reverse easier
AD_ENC_TYPE_TO_KRB5_ENC_TYPE_MAP = {
    ADEncryptionType.AES256_CTS_HMAC_SHA1_96: "aes256-cts-hmac-sha1-96",
    ADEncryptionType.AES128_CTS_HMAC_SHA1_96: "aes128-cts-hmac-sha1-96",
    ADEncryptionType.RC4_HMAC: "arcfour-hmac",
}
# the reverse of the above
KRB5_ENC_TYPE_TO_AD_ENC_TYPE_MAP = {value: key for key, value in AD_ENC_TYPE_TO_KRB5_ENC_TYPE_MAP.items()}


# used for reading in keytabs. AD only supports a small handful of these (from the map above)
# but we include all encryption types here for completeness.
KRB5_ENC_TYPE_VALUE_TO_ENC_TYPE_MAP = {
    26: "camellia256-cts-cmac",  # RFC6803
    25: "camellia128-cts-cmac",  # RFC6803
    24: "arcfour-hmac",  # RFC8429. This is arcfour-hmac exportable. but the exportable doesn't matter here, just treat it as arcfour for user views
    23: "arcfour-hmac",  # RFC8429
    # 21 and 22 are unassigned
    20: "aes256-cts-hmac-sha384-192",  # RFC8009
    19: "aes128-cts-hmac-sha256-128",  # RFC8009
    18: "aes256-cts-hmac-sha1-96",  # RFC3962
    17: "aes128-cts-hmac-sha1-96",  # RFC3962
    16: "des3-cbc-sha1",  # RFC8429

    # 9 through 15 should never actually be used in NFS or SMB kerberos. but it doesn't hurt to be
    # future-proof
    15: "des-ede3-cbc-Env-OID",  # RFC4556
    14: "rsaES-OAEP-ENV-OID",  # RFC4556
    13: "rsaEncryption-EnvOID",  # RFC4556
    12: "arc2-cbc-EnvOID",  # RFC4556
    11: "sha1WithRSAEncryption-CmsOID",  # RFC4556
    10: "md5WithRSAEncryption-CmsOID",  # RFC4556
    9: "dsaWithSHA1-CmsOID",  # RFC4556

    8: "des-hmac-sha1",
    7: "des3-cbc-sha1",  # RFC8429
    # 6 is reserved RFC3961
    5: "des3-cbc-md5",  # RFC8429
    # 4 is reserved RFC3961
    3: "des-cbc-md5",  # RFC6649
    2: "des-cbc-md4",  # RFC6649
    1: "des-cbc-crc",  # RFC6649
}
# this is a reverse of the above
KRB5_ENC_TYPE_TO_ENC_TYPE_VALUE_MAP = {value: key for key, value in KRB5_ENC_TYPE_VALUE_TO_ENC_TYPE_MAP.items()}
NAME_TYPE_VALUE_TO_NAME_TYPE_MAP = {
    0: "KRB5_NT_UNKNOWN",
    1: "KRB5_NT_PRINCIPAL",
    2: "KRB5_NT_SRV_INST",
    5: "KRB5_NT_UID",
}
DEFAULT_UNKNOWN_NAME_TYPE = 0
AD_DEFAULT_NAME_TYPE = 1

DEFAULT_KRB5_KEYTAB_FILE_LOCATION = '/etc/krb5.keytab'



import copy
import socket
import ssl

from ms_active_directory import logging_utils

from ldap3 import (
    BASE,
    Connection,
    SUBTREE,
)
from typing import List

import ms_active_directory.environment.constants as constants
import ms_active_directory.environment.ldap.ldap_format_utils as ldap_utils
import ms_active_directory.environment.ldap.ldap_constants as ldap_constants
import ms_active_directory.environment.security.security_config_utils as security_utils
import ms_active_directory.environment.security.security_config_constants as security_constants

from ms_active_directory.core.ad_computer import ADComputer
from ms_active_directory.core.ad_users_and_groups import ADGroup, ADUser, ADObject
from ms_active_directory.exceptions import (
    DomainSearchException,
    DuplicateNameException,
    MembershipModificationException,
    MembershipModificationRollbackException,
    ObjectCreationException,
)

logger = logging_utils.get_logger()


class ADSession:

    def __init__(self, ldap_connection: Connection, domain):
        self.ldap_connection = ldap_connection
        self.domain = domain
        self.domain_dns_name = self.domain.get_domain_dns_name()
        self.domain_search_base = ldap_utils.construct_ldap_base_dn_from_domain(self.domain_dns_name)
        # when checking whether we can create something within the domain, we check for sAMAccountName conflicts
        # and other conflicts depending on what's being created. when doing that validation we should always use
        # the plain domain base, whereas users might confine self.domain_search_base later to perform lookups
        # in a subsection of the domain
        self._domain_validation_search_base = self.domain_search_base

    def is_authenticated(self):
        """ Returns if the session is currently authenticated """
        return self.ldap_connection.bound

    def is_encrypted(self):
        """ Returns if the session's connection is encrypted """
        return self.ldap_connection.tls_started or self.ldap_connection.server.ssl

    def is_open(self):
        """ Returns if the session's connection is currently open """
        return not self.ldap_connection.closed

    def is_thread_safe(self):
        """ Returns if the session's connection is thread-safe """
        return self.ldap_connection.strategy.thread_safe

    def get_ldap_connection(self):
        """ Returns the LDAP connection that this session uses for communication.
        This is particularly useful if a user wants to make complex LDAP queries or perform
        operations that are not supported by the ADSession object, and is willing to craft
        them and parse results themselves.
        """
        return self.ldap_connection

    def get_current_server_uri(self):
        """ Returns the URI of the server that this session is currently communicating with """
        return self.ldap_connection.server.name

    def get_domain(self):
        """ Returns the domain that this session is connected to """
        return self.domain

    def get_domain_dns_name(self):
        """ Returns the domain that this session is connected to """
        return self.domain_dns_name

    def get_domain_search_base(self):
        """ Returns the LDAP search base used for all 'find' functions as the search base """
        return self.domain_search_base

    def set_domain_search_base(self, search_base: str):
        """ Set the search base to use for 'find' queries within the domain made by this session.
        This can be used to confine our search to a sub-container within the domain. This can improve
        the performance of lookups, avoid permissioning issues, and remove issues around duplicate
        records with the same common name.
        """
        # an empty string search base is a special search base and shouldn't be normalized
        if search_base != '':
            normalized_rdn = ldap_utils.normalize_object_location_in_domain(search_base,
                                                                            self.domain_dns_name)
            # always use the base 'DC=...,DC=...' for our normalization
            search_base = normalized_rdn + ',' + self._domain_validation_search_base
        self.domain_search_base = search_base

    def dn_exists_in_domain(self, distinguished_name: str):
        """ Check if a distinguished name exists within the domain, regardless of what it is.
        :param distinguished_name: Either a relative distinguished name or full distinguished name
                                   to search for within the domain.
        :returns: True if the distinguished name exists within the domain.
        """
        # since the distinguished name may be a relative distinguished name or complete one,
        # normalize it
        normalized_rdn = ldap_utils.normalize_object_location_in_domain(distinguished_name,
                                                                        self.domain_dns_name)
        search_dn = normalized_rdn + ',' + self.domain_search_base
        # search returns True if it finds anything. we only need to find one object before stopping
        res = self.ldap_connection.search(search_base=search_dn,
                                          search_filter=ldap_constants.FIND_ANYTHING_FILTER,
                                          search_scope=BASE,
                                          size_limit=1)
        exists, _, _, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        return exists

    def object_exists_in_domain_with_attribute(self, attr: str, unescaped_value: str):
        """ Check if any objects exist in the domain with a given attribute. Returns True if so, False otherwise.
        :param attr: The LDAP attribute to examine in the search.
        :param unescaped_value: The value of the attribute that we're looking for, in its raw form.
        :returns: True if any objects exist in the domain with the attribute specified equal to the value.
        """
        if ldap_utils.is_dn(unescaped_value):
            value = ldap_utils.escape_dn_for_filter(unescaped_value)
            logger.debug('Escaped value %s of %s as distinguished name to be %s', unescaped_value, attr, value)
        else:
            value = ldap_utils.escape_generic_filter_value(unescaped_value)
            logger.debug('Escaped value %s of %s as generic LDAP value to be %s', unescaped_value, attr, value)
        ldap_filter = '({}={})'.format(attr, value)
        # search returns True if it finds anything. we only need to find one object before stopping
        res = self.ldap_connection.search(search_base=self.domain_search_base,
                                          search_filter=ldap_filter,
                                          search_scope=SUBTREE,
                                          attributes=[attr],
                                          size_limit=1)
        _, _, response, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        real_entities = ldap_utils.remove_ad_search_refs(response)
        return len(real_entities) > 0

    def _create_object(self, object_dn: str, object_classes: List[str], account_attr_dict: dict):
        if self.dn_exists_in_domain(object_dn):
            raise ObjectCreationException('An object already exists within the domain with distinguished name {} - '
                                          'please remove it or change the attributes specified such that a different '
                                          'distinguished name is created.'.format(object_dn))
        res = self.ldap_connection.add(object_dn, object_classes, account_attr_dict)
        # TODO: returning the actual response is probably a good idea so we can do something with it.
        # doing more with it in case of error would be a good idea too
        success, result, _, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if success:
            return success
        # don't include attributes in the exception because a password could be there and it could get logged.
        raise ObjectCreationException('An exception was encountered creating an object with distinguished name {} '
                                      'and object classes {}. LDAP result: {}'.format(object_dn, object_classes, result))

    def create_computer(self, computer_name: str, computer_location: str=None, computer_password: str=None,
                        encryption_types: List[security_constants.ADEncryptionType]=None, hostnames: List[str]=None,
                        services: List[str]=None, supports_legacy_behavior: bool=False,
                        **additional_account_attributes):
        """ Use the session to create a computer in the domain and return a computer object.
        :param computer_name: The common name of the computer to create in the AD domain. This
                              will be used to determine the sAMAccountName, and if no hostnames
                              are specified then this will be used to determine the hostnames for
                              the computer.
        :param computer_location: The distinguished name of the location within the domain where
                                  the computer will be created. It may be a relative distinguished
                                  name (not including the domain component) or a full distinguished
                                  name.  If not specified, defaults to CN=Computers which is
                                  standard for Active Directory.
        :param computer_password: The password to be set for the computer. This is particularly
                                  useful to specify if the computer will be shared across multiple
                                  applications or devices, or if pre-creating a computer for another
                                  application to use. If not specified, a random 120 character
                                  password will be generated.
        :param encryption_types: The encryption types to set as supported on the computer in AD.
                                 These will also be used to generate kerberos keys for the computer.
                                 If not specified, defaults to [aes256-cts-hmac-sha1-96].
        :param hostnames: The hostnames to use for configuring the service principal names of the
                          computer. These may be short hostnames or fully qualified domain names.
                          If not specified, defaults to the "computer_name" as a short hostname and
                          "computer_name.domain" as a fully qualified domain name.
        :param services: The services to enable on each hostname, which will be used with hostnames
                         to generate the computer's service principal names. If not specified,
                         defaults to ["HOST"] which is standard for Active Directory.
        :param supports_legacy_behavior: Does the computer being created support legacy behavior such
                                         as NTLM authentication or UNC path addressing from older windows
                                         clients? Defaults to False. Impacts the restrictions on
                                         computer naming.
        :param additional_account_attributes: Additional LDAP attributes to set on the account and their
                                              values. This is used to support power users setting arbitrary
                                              attributes, such as "userCertificate" to set the certificate
                                              for a computer that will use mutual TLS for EXTERNAL SASL auth.
                                              This also allows overriding of some values that are not explicit
                                              keyword arguments in order to avoid overcomplication, since most
                                              people won't set them (e.g. userAccountControl).
        """
        logger.debug('Request to create computer in domain %s with the following attributes: computer_name=%s, '
                     'computer_location=%s encryption_types=%s hostnames=%s services=%s supports_legacy_behavior=%s '
                     'number of additional attributes specified: %s', self.domain_dns_name, computer_name,
                     computer_location, encryption_types, hostnames, services, supports_legacy_behavior,
                     len(additional_account_attributes))
        # validate our computer name and then determine our sAMAccountName
        computer_name = ldap_utils.validate_and_normalize_computer_name(computer_name, supports_legacy_behavior)
        samaccount_name = computer_name + '$'

        if self.object_exists_in_domain_with_attribute(ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME, samaccount_name):
            raise ObjectCreationException('An object already exists with sAMAccountName {} so a computer may not be '
                                          'created with the name {}'.format(samaccount_name, computer_name))

        # get or normalize our computer location. the end format is as a relative distinguished name
        if computer_location is None:
            computer_location = ldap_constants.DEFAULT_COMPUTER_LOCATION
        else:
            computer_location = ldap_utils.normalize_object_location_in_domain(computer_location,
                                                                               self.domain_dns_name)
        if not self.dn_exists_in_domain(computer_location):
            raise ObjectCreationException('The computer location {} cannot be found in the domain.'
                                          .format(computer_location))
        # now we can build our full object distinguished name
        computer_dn = ldap_utils.construct_object_distinguished_name(computer_name, computer_location,
                                                                     self.domain_dns_name)
        if self.dn_exists_in_domain(computer_dn):
            raise ObjectCreationException('There exists an object in the domain with distinguished name {} and so a '
                                          'computer may not be created in the domain with name {} in location {}. '
                                          'Please use a different name or location.'
                                          .format(computer_dn, computer_name, computer_location))

        # generate a password if needed and encode it
        if computer_password is None:
            computer_password = security_utils.generate_random_ad_password()
        encoded_pw = security_utils.encode_password(computer_password)

        # normalize encryption type values and convert it to the encoded bitstring
        if encryption_types is None:
            encryption_types = [security_constants.ADEncryptionType.AES256_CTS_HMAC_SHA1_96]
        else:
            encryption_types = security_utils.normalize_encryption_type_list(encryption_types)
        encoded_enc_type_value = security_utils.get_supported_encryption_types_value(encryption_types)

        if hostnames is None:
            hostnames = ldap_utils.construct_default_hostnames_for_computer(computer_name, self.domain_dns_name)
        if services is None:
            services = ldap_constants.DEFAULT_COMPUTER_SERVICES
        spns = ldap_utils.construct_service_principal_names(services, hostnames)
        for spn in spns:
            if self.object_exists_in_domain_with_attribute(ldap_constants.AD_ATTRIBUTE_SERVICE_PRINCIPAL_NAMES, spn):
                raise ObjectCreationException('An object exists in the domain with service principal name {} and so '
                                              'creating a computer with the hostnames ({}) and services ({}) in use '
                                              'will cause undefined, conflicting behavior during lookups. Please '
                                              'specify different hostnames or services, or a different computer name '
                                              'if hostnames are not being explicitly set.'
                                              .format(spn, ', '.join(hostnames), ', '.join(services)))

        computer_attributes = {
            ldap_constants.AD_ATTRIBUTE_USER_ACCOUNT_CONTROL: ldap_constants.COMPUTER_ACCESS_CONTROL_VAL,
            ldap_constants.AD_ATTRIBUTE_PASSWORD: encoded_pw,
            ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME: samaccount_name,
            ldap_constants.AD_ATTRIBUTE_ENCRYPTION_TYPES: encoded_enc_type_value,
            ldap_constants.AD_ATTRIBUTE_SERVICE_PRINCIPAL_NAMES: spns,
            ldap_constants.AD_ATTRIBUTE_DNS_HOST_NAME: hostnames[0],
        }
        if len(hostnames) > 1:
            computer_attributes[ldap_constants.AD_ATTRIBUTE_ADDITIONAL_DNS_HOST_NAME] = hostnames[1:]
        # don't include additional attributes for logging in case they're sensitive
        loggable_attributes = copy.deepcopy(computer_attributes)
        del loggable_attributes[ldap_constants.AD_ATTRIBUTE_PASSWORD]
        logger.info('Attempting to create computer in domain %s with the following LDAP attributes: %s and %s additional attributes',
                    loggable_attributes, len(additional_account_attributes))

        # add in our additional account attributes at the end so they can override anything we set here
        computer_attributes.update(additional_account_attributes)

        self._create_object(computer_dn, ldap_constants.OBJECT_CLASSES_FOR_COMPUTER, computer_attributes)
        return ADComputer(samaccount_name, self.domain, computer_location, computer_password, spns,
                          encryption_types)

    def take_over_existing_computer(self, computer_name: str):
        """ Use the session to take over a computer in the domain and return a computer object.
        This resets the computer's password so that nobody else can impersonate it, and reads
        the computer's attributes in order to create a computer object and return it.
        :param computer_name: The common name or sAMAccountName of the computer to find in the AD domain.
                              If it appears to be a common name, not ending in $, a sAMAccountName will
                              be derived to search for. If that cannot be found, then a search will be
                              done for this as a common name. If no unique computer can be found with that
                              search, then an exception will be raised.
        """
        raise NotImplementedError()

    # FUNCTIONS FOR FINDING DOMAIN INFORMATION

    def is_domain_close_in_time_to_localhost(self, allowed_drift_seconds=None):
        """ Get whether the domain time is close to the current local time.
        Just calls the parent domain function and returns that. This is included here for completeness.
        """
        return self.domain.is_close_in_time_to_localhost(self.ldap_connection, allowed_drift_seconds)

    def find_certificate_authorities_for_domain(self, pem_format: bool=True):
        """ Attempt to discover the CAs within the domain and return info on their certificates.
        If a session was first established using an IP address or blind trust TLS, but we want to bootstrap our
        sessions to establish stronger trust, or write the CA certificates to a local truststore for other
        non-LDAP applications to use (e.g. establishing roots of trust for https or syslog over TLS), then it's
        helpful to grab the certificate authorities in the domain and their signing certificates.
        Not all domains run certificate authorities; some use public CAs or get certs from other PKI being run,
        so this isn't useful for everyone. But a lot of people do run CAs in their AD domains, and this is useful
        for them.

        :param pem_format: If True, return the certificates as strings in PEM format. Otherwise, return the
                           certificates as bytestrings in DER format. Defaults to True.
        :returns: A list of either PEM-formatted certificate strings or DER-formatted certificate byte strings,
                  representing the CA certificates of the CAs within the domain.
        """
        ca_filter = '({}={})'.format(ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                                     ldap_constants.CERTIFICATE_AUTHORITY_OBJECT_CLASS)
        search_loc = '{},{}'.format(ldap_constants.DOMAIN_WIDE_CONFIGURATIONS_CONTAINER,
                                    self.domain_search_base)
        res = self.ldap_connection.search(search_base=search_loc, search_filter=ca_filter, search_scope=SUBTREE,
                                          attributes=[ldap_constants.AD_ATTRIBUTE_CA_CERT])
        success, _, entities, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search domain for Certificate Authorities')
        entities = ldap_utils.remove_ad_search_refs(entities)
        logger.info('Found %s CAs within the domain', len(entities))
        ca_certs_der_fmt = []
        for entity in entities:
            # the bytes are returned in a 1-item list
            ca_cert_der_bytes = entity['attributes'][ldap_constants.AD_ATTRIBUTE_CA_CERT][0]
            ca_certs_der_fmt.append(ca_cert_der_bytes)

        if pem_format:
            return [ssl.DER_cert_to_PEM_cert(cert_bytes) for cert_bytes in ca_certs_der_fmt]
        return ca_certs_der_fmt

    def find_current_time_for_domain(self):
        """ Get the current time for the domain as a datetime object.
        Just calls the parent domain function and returns that. This is included here for completeness.
        :returns: A datetime object representing the current time in the domain.
        """
        return self.domain.find_current_time(self.ldap_connection)

    def find_dns_servers_for_domain(self):
        """ Attempt to discover the DNS servers within the domain and return info on them.
        If a session was first established using an IP address or blind trust TLS, but we want to bootstrap our
        sessions to use kerberos or TLS backed by CA certificates, we need proper DNS configured. For private
        domains (e.g. in a datacenter), we may run DNS servers within the domain. This function discovers
        computers with a "DNS/" service principal name, tries to look up IP addresses for them, and then
        returns that information.
        This won't always be useful, as DNS isn't always part of the AD domain, but it can help if we're bootstrapping
        a computer with manufacturer configurations to use the AD domain for everything based on a minimal starting
        configuration.

        :returns: A dictionary mapping DNS hostnames of DNS servers to IP addresses. The hostnames are provided in case
                  a caller is configuring DNS-over-TLS. If no IP address can be resolved for a hostname, it will map to
                  a None value.
                  https://datatracker.ietf.org/doc/html/rfc8310
        """
        dns_computer_filter = '(&({}={})({}={}))'.format(ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                                                         ldap_constants.COMPUTER_OBJECT_CLASS,
                                                         ldap_constants.AD_ATTRIBUTE_SERVICE_PRINCIPAL_NAMES,
                                                         ldap_constants.DNS_SERVICE_FILTER)
        # search the whole domain and grab the dns hostnames of the computers found
        res = self.ldap_connection.search(search_base=self.domain_search_base, search_filter=dns_computer_filter,
                                          search_scope=SUBTREE, attributes=[ldap_constants.AD_ATTRIBUTE_DNS_HOST_NAME])
        success, _, entities, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search domain for Computers hosting DNS services')
        entities = ldap_utils.remove_ad_search_refs(entities)
        logger.info('Found %s computers that host DNS services within the domain', len(entities))

        dns_results = {}
        for entity in entities:
            hostname = entity['attributes'][ldap_constants.AD_ATTRIBUTE_DNS_HOST_NAME]
            ip_addr = None
            try:
                ip_addr = socket.gethostbyname(hostname)
            except:
                logger.warning('Unable to resolve ip address for %s', hostname)

            if hostname in dns_results:
                logger.warning('Multiple computer entities exist in the domain serving DNS services with the DNS '
                               'hostname %s - this may indicate a domain misconfiguration issue.', hostname)
            dns_results[hostname] = ip_addr
        return dns_results

    def find_forest_schema_version(self):
        """ Attempt to determine the version of Windows Server set in the forest's schema.
        returns: An Enum of type ADVersion indicating the schema version.
        """
        search_loc = '{},{}'.format(ldap_constants.DOMAIN_CONTROLLER_SCHEMA_VERSION_SEARCH_CONTAINER,
                                    self.domain_search_base)
        res = self.ldap_connection.search(search_base=search_loc, search_filter=ldap_constants.FIND_ANYTHING_FILTER,
                                          search_scope=BASE, attributes=[ldap_constants.AD_ATTRIBUTE_GET_ALL_NON_VIRTUAL_ATTRS])
        success, _, entities, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search domain for schema.')
        entities = ldap_utils.remove_ad_search_refs(entities)
        if len(entities) == 0:
            raise DomainSearchException('The forest schema could not be found when searching the domain.')
        schema = entities[0]
        ad_schema_ver = schema['attributes'][ldap_constants.AD_SCHEMA_VERSION]
        return constants.ADVersion.get_version_from_schema_number(ad_schema_ver)

    def find_functional_level_for_domain(self):
        """ Attempt to discover the functional level of the domain and return it.
        This will indicate if the domain is operating at the level of a 2008, 2012R2, 2016, etc. domain.
        The functional level of a domain influences what functionality exists (e.g. 2003 cannot issue AES keys,
        2012 cannot use many TLS ciphers introduced with TLS1.3) and so it can be useful for determining what
        to do.
        :returns: An Enum of type ADFunctionalLevel indicating the functional level.
        """
        return self.domain.find_functional_level(self.ldap_connection)

    def find_supported_sasl_mechanisms_for_domain(self):
        """ Attempt to discover the SASL mechanisms supported by the domain and return them.
        This just builds upon the functionality that the domain has for this, as you don't need
        to be authenticated as anything other than anonymous to read this information (since it's
        often used to figure out how to authenticate).
        This is included in the session object for completeness.
        """
        return self.domain.find_supported_sasl_mechanisms(self.ldap_connection)

    # FUNCTIONS FOR FINDING USERS AND GROUPS

    def _find_ad_objects_and_attrs(self, search_base: str, search_filter:str, search_scope: str,
                                   attributes: List[str], size_limit: int, return_type):
        """ A helper function for common search and result parsing logic in other find functions for users
        and groups
        """
        attrs = self._figure_out_search_attributes_for_user_or_group(attributes)
        res = self.ldap_connection.search(search_base=search_base,
                                          search_filter=search_filter,
                                          search_scope=search_scope,
                                          size_limit=size_limit,
                                          attributes=attrs)
        _, _, resp, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        resp = ldap_utils.remove_ad_search_refs(resp)
        if not resp:
            return []

        results = []
        for entry in resp:
            entry_attributes = entry['attributes']
            obj = return_type(entry['dn'], entry_attributes, self.domain)
            results.append(obj)
        return results

    def find_objects_with_attribute(self, attribute_name: str, attribute_value, attributes_to_lookup: List[str]=None,
                                    object_class: str=None, return_type=None):
        """ Find all AD objects that possess the specified attribute with the specified value and return them.

        :param attribute_name: The LDAP name of the attribute to be used in the search.
        :param attribute_value: The value that returned objects should possess for the attribute.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the groups' name and object class attributes will be queried.
        :param object_class: Optional. The object class to filter on when searching. Defaults to 'top' which will
                             include all objects in AD.
        :param return_type: Optional. The class to use to represent the returned objects. Defaults to ADObject.
                            If a generic search is being done, or an object class is used that is not yet supported
                            by this library, using ADObject is recommended.
        :returns: a list of ADGroup objects representing groups with the specified value for the specified attribute.
        """
        if return_type is None:
            return_type = ADObject
        if object_class is None:
            # top is the top level object class that all objects possess. it's as generic as you can get
            object_class = ldap_constants.TOP_OBJECT_CLASS

        # if our attribute value looks like a DN, escape it like one. otherwise, escape normally
        if ldap_utils.is_dn(attribute_value):
            escaped_val = ldap_utils.escape_dn_for_filter(attribute_value)
        else:
            escaped_val = ldap_utils.escape_generic_filter_value(attribute_value)
        search_filter = ('(&({obj_class_attr}={obj_class})({attr}={attr_val}))'
                         .format(obj_class_attr=ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                                 obj_class=object_class, attr=attribute_name,
                                 attr_val=escaped_val))
        # at a minimum, always look up the thing we're filtering on
        if attributes_to_lookup is None:
            attributes_to_lookup = [attribute_name]
        if attribute_name not in attributes_to_lookup:
            attributes_to_lookup.append(attribute_name)
        # a size limit of 0 means unlimited
        res = self._find_ad_objects_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                              attributes_to_lookup, 0, return_type)
        logger.info('%s %s objects found with %s value %s', len(res), object_class, attribute_name, attribute_value)
        return res

    def find_groups_by_attribute(self, attribute_name: str, attribute_value, attributes_to_lookup: List[str]=None):
        """ Find all groups that possess the specified attribute with the specified value, and return a list of ADGroup
        objects.

        :param attribute_name: The LDAP name of the attribute to be used in the search.
        :param attribute_value: The value that returned groups should possess for the attribute.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the groups' name and object class attributes will be queried.
        :returns: a list of ADGroup objects representing groups with the specified value for the specified attribute.
        """
        return self.find_objects_with_attribute(attribute_name, attribute_value, attributes_to_lookup,
                                                ldap_constants.GROUP_OBJECT_CLASS, ADGroup)

    def find_groups_by_common_name(self, group_name: str, attributes_to_lookup: List[str]=None):
        """ Find all groups with a given common name and return a list of ADGroup objects.
        This is particularly useful when you have multiple groups with the same name in different OUs
        as a result of a migration, and want to find them so you can combine them.

        :param group_name: The common name of the group(s) to be looked up.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the groups' name and object class attributes will be queried.
        :returns: a list of ADGroup objects representing groups with the specified common name.
        """
        # build a compound filter for users with this common name
        search_filter = '(&({cn_attr}={cn}){type_filter})'.format(cn_attr=ldap_constants.AD_ATTRIBUTE_COMMON_NAME,
                                                                  cn=group_name,
                                                                  type_filter=ldap_constants.FIND_GROUP_FILTER)
        # a size limit of 0 means unlimited
        res = self._find_ad_objects_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                              attributes_to_lookup, 0, ADGroup)
        logger.info('%s groups found with common name %s', len(res), group_name)
        return res

    def find_group_by_distinguished_name(self, group_dn: str, attributes_to_lookup: List[str]=None):
        """ Find a group in AD based on a specified distinguished name and return it along with any
        requested attributes.
        :param group_dn: The distinguished name of the group.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the group's name and object class attributes will be queried.
        :returns: an ADGroup object or None if the group does not exist.
        """
        # since the distinguished name may be a relative distinguished name or complete one,
        # normalize it
        normalized_rdn = ldap_utils.normalize_object_location_in_domain(group_dn,
                                                                        self.domain_dns_name)
        search_dn = normalized_rdn + ',' + self.domain_search_base
        res = self._find_ad_objects_and_attrs(search_dn, ldap_constants.FIND_GROUP_FILTER, BASE,
                                              attributes_to_lookup, 1, ADGroup)
        if not res:
            return None
        return res[0]

    def find_group_by_sam_name(self, group_name: str, attributes_to_lookup: List[str]=None):
        """ Find a Group in AD based on a specified sAMAccountName name and return it along with any
        requested attributes.
        :param group_name: The sAMAccountName name of the group.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the group's name and object class attributes will be queried.
        :returns: an ADGroup object or None if the group does not exist.
        """
        # build a compound filter for users with this samaccountname
        search_filter = '(&({sam_attr}={sam_name}){type_filter})'.format(sam_attr=ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME,
                                                                         sam_name=group_name,
                                                                         type_filter=ldap_constants.FIND_GROUP_FILTER)
        res = self._find_ad_objects_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                              attributes_to_lookup, 1, ADGroup)
        if not res:
            return None
        return res[0]

    def find_group_by_name(self, group_name: str, attributes_to_lookup: List[str]=None):
        """ Find a Group in AD based on a provided name.
        This function takes in a generic name which can be either a distinguished name, a common name, or a
        sAMAccountName, and tries to find a unique group identified by it and return information on the group.
        :param group_name: The name of the group, which may be a DN, common name, or sAMAccountName.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the group's name and object class attributes will be queried.
        :returns: an ADGroup object or None if the group does not exist.
        :raises: a DuplicateNameException if more than one entry exists with this name.
        """
        return self._find_by_name_common(group_name, attributes_to_lookup, is_user=False)

    def find_users_by_attribute(self, attribute_name: str, attribute_value, attributes_to_lookup: List[str]=None):
        """ Find all users that possess the specified attribute with the specified value, and return a list of ADUser
        objects.

        :param attribute_name: The LDAP name of the attribute to be used in the search.
        :param attribute_value: The value that returned groups should possess for the attribute.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the users. Regardless of
                                     what's specified, the users' name and object class attributes will be queried.
        :returns: a list of ADUser objects representing groups with the specified value for the specified attribute.
        """
        return self.find_objects_with_attribute(attribute_name, attribute_value, attributes_to_lookup,
                                                ldap_constants.USER_OBJECT_CLASS, ADUser)

    def find_users_by_common_name(self, user_name: str, attributes_to_lookup: List[str]=None):
        """ Find all users with a given common name and return a list of ADUser objects.
        This is particularly useful when you have multiple users with the same name in different OUs
        as a result of a migration, and want to find them so you can combine them.

        :param user_name: The common name of the user(s) to be looked up.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the users. Regardless of
                                     what's specified, the users' name and object class attributes will be queried.
        :returns: a list of ADUser objects representing users with the specified common name.
        """
        # build a compound filter for users with this common name
        search_filter = '(&({cn_attr}={cn}){type_filter})'.format(cn_attr=ldap_constants.AD_ATTRIBUTE_COMMON_NAME,
                                                                  cn=user_name,
                                                                  type_filter=ldap_constants.FIND_USER_FILTER)
        # a size limit of 0 means unlimited
        res = self._find_ad_objects_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                              attributes_to_lookup, 0, ADUser)
        logger.info('%s users found with common name %s', len(res), user_name)
        return res

    def find_user_by_distinguished_name(self, user_dn: str, attributes_to_lookup: List[str]=None):
        """ Find a User in AD based on a specified distinguished name and return it along with any
        requested attributes.
        :param user_dn: The distinguished name of the user.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the user. Regardless of
                                     what's specified, the user's name and object class attributes will be queried.
        :returns: an ADUser object or None if the user does not exist.
        """
        # since the distinguished name may be a relative distinguished name or complete one,
        # normalize it
        normalized_rdn = ldap_utils.normalize_object_location_in_domain(user_dn,
                                                                        self.domain_dns_name)
        search_dn = normalized_rdn + ',' + self.domain_search_base
        res = self._find_ad_objects_and_attrs(search_dn, ldap_constants.FIND_USER_FILTER, BASE,
                                              attributes_to_lookup, 1, ADUser)
        if not res:
            return None
        return res[0]

    def find_user_by_sam_name(self, user_name, attributes_to_lookup: List[str]=None):
        """ Find a User in AD based on a specified sAMAccountName name and return it along with any
        requested attributes.
        :param user_name: The sAMAccountName name of the user.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the user. Regardless of
                                     what's specified, the user's name and object class attributes will be queried.
        :returns: an ADUser object or None if the user does not exist.
        """
        # build a compound filter for users with this samaccountname
        search_filter = '(&({sam_attr}={sam_name}){type_filter})'.format(sam_attr=ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME,
                                                                         sam_name=user_name,
                                                                         type_filter=ldap_constants.FIND_USER_FILTER)
        res = self._find_ad_objects_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                              attributes_to_lookup, 1, ADUser)
        if not res:
            return None
        return res[0]

    def find_user_by_name(self, user_name: str, attributes_to_lookup: List[str]=None):
        """ Find a User in AD based on a provided name.
        This function takes in a generic name which can be either a distinguished name, a common name, or a
        sAMAccountName, and tries to find a unique user identified by it and return information on the user.
        :param user_name: The name of the user, which may be a DN, common name, or sAMAccountName.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the user. Regardless of
                                     what's specified, the user's name and object class attributes will be queried.
        :returns: an ADUser object or None if the user does not exist.
        :raises: a DuplicateNameException if more than one entry exists with this name.
        """
        return self._find_by_name_common(user_name, attributes_to_lookup, is_user=True)

    def _find_by_name_common(self, name: str, attributes_to_lookup: List[str], is_user: bool=True):
        """ A helper function to find things by a unique name.
        Depending on the type of object, it first checks if the name is a DN. We then perform
        the appropriate lookup by DN if it is. Otherwise, we first try a lookup by sAMAccountName;
        if that works we return that. If it doesn't, we try a lookup by common name. If that finds
        1 result, we return it. Otherwise, we raise a duplicate name error.
        """
        is_dn = ldap_utils.is_dn(name)
        dn_lookup_func = self.find_group_by_distinguished_name
        sam_lookup_func = self.find_group_by_sam_name
        cn_lookup_func = self.find_groups_by_common_name
        if is_user:
            dn_lookup_func = self.find_user_by_distinguished_name
            sam_lookup_func = self.find_user_by_sam_name
            cn_lookup_func = self.find_users_by_common_name

        if is_dn:
            return dn_lookup_func(name, attributes_to_lookup)
        res = sam_lookup_func(name, attributes_to_lookup)
        if res:
            return res
        result_list = cn_lookup_func(name, attributes_to_lookup)
        if not result_list:
            return None
        if len(result_list) > 1:
            insert = 'user' if is_user else 'group'
            raise DuplicateNameException('Multiple {}s found with name "{}". Please either repeat the search '
                                         'using a distinguished name or sAMAccountName, or adjust the session '
                                         'search base using set_domain_search_base to limit searches such that only '
                                         'one result is found. Alternatively you may perform a lookup by common name '
                                         'and select which {} entry you want to use from multiple.'
                                         .format(insert, name, insert))
        return result_list[0]

    def _figure_out_search_attributes_for_user_or_group(self, attributes_to_lookup):
        """ There's some attributes we'll always get for users and groups, whether callers requested them or not.
        This combines those with any requested attributes
        """
        base_attrs = {ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME, ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                      ldap_constants.AD_ATTRIBUTE_COMMON_NAME}
        if attributes_to_lookup:
            base_attrs.update(set(attributes_to_lookup))
        # sort for reproducibility in eventual testing
        return sorted(list(base_attrs))

    # FUNCTIONS FOR FINDING MEMBERSHIP INFORMATION

    def find_groups_for_entities(self, entities: List, attributes_to_lookup: List[str]=None,
                                 lookup_by_name_fn: callable=None):
        """ Find the parent groups for all of the entities in a List.
        These entities may be users, groups, or anything really because Active Directory uses the "groupOfNames" style
        membership tracking, so all group members are just represented as distinguished names regardless of type.
        If the elements of entities are strings and are not distinguished names, then lookup_by_name_fn will be used
        to look up the appropriate ADObject for the entity and get its distinguished name.

        The parent groups of all the entities will then be queried, and the attributes specified will be looked up
        (if any). A dictionary mapping the original entities to lists of ADGroup objects will be returned.

        :param entities: A list of either ADObject objects or strings. These represent the objects whose parent groups
                         are being queried.
        :param attributes_to_lookup: A list of LDAP attributes to query about the parent groups, in addition to the
                                     default ones queries. Optional.
        :param lookup_by_name_fn: An optional function to call to map entities to ADObjects when the members of entities
                                  are strings that are not LDAP distinguished names.
        :returns: A dictionary mapping input entities to lists of ADGroup object representing their parent groups.
        :raises: a DuplicateNameException if an entity name is specified and more than one entry exists with the name.
        :raises: InvalidLdapParameterException if any non-string non-ADObject types are found in entities, or if any
                 non-distinguished name strings are specified.
        """
        # make a map of entity distinguished names to entities passed in. we'll use this when constructing
        # our return dictionary as well
        entity_dns = ldap_utils.normalize_entities_to_entity_dns(entities, lookup_by_name_fn)

        filter_pieces = []
        for entity_dn in entity_dns:
            filter_piece = '({}={})'.format(ldap_constants.AD_ATTRIBUTE_MEMBER,
                                            ldap_utils.escape_dn_for_filter(entity_dn))
            filter_pieces.append(filter_piece)
        all_member_filters = ''.join(filter_pieces)
        # this filter can get really really big, because DNs are long and we can have a lot of entities.
        # AD does have a limit on request size; that limit is huge by default though - 10MB - and can be raised
        # https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2003/cc755809(v=ws.10)
        chain_filter = ('(&({obj_cls_attr}={obj_cls})(|{member_filters}))'
                        .format(obj_cls_attr=ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                                obj_cls=ldap_constants.GROUP_OBJECT_CLASS,
                                member_filters=all_member_filters))
        filter_length = len(chain_filter)
        logger.info('Querying parent groups of %s entitites using a filter of length %s',
                    len(entities), filter_length)
        # if our filter is more than 95% of the default limit, log at a warning level to make debugging issues easier
        if filter_length > 10000000:
            logger.warning('Filter length exceeds 95% of the DEFAULT active directory limit on request size, so this '
                           'request may be rejected, in which case you should re-attempt with fewer entities in '
                           'batches', filter_length)

        # make sure member is included in our results
        if attributes_to_lookup is None or attributes_to_lookup is []:
            attributes_to_lookup = []
        if ldap_constants.AD_ATTRIBUTE_MEMBER not in attributes_to_lookup:
            attributes_to_lookup.append(ldap_constants.AD_ATTRIBUTE_MEMBER)
        # look up our results
        results = self._find_ad_objects_and_attrs(self.domain_search_base, chain_filter, SUBTREE,
                                                  attributes_to_lookup, 0, ADGroup)
        logger.info('Found %s unique parent groups across all %s entities', len(results), len(entities))
        mapping_dict = {}
        for entity in entities:
            mapping_dict[entity] = []

        # go over our results and for each one, figure out which input entities have the group as a parent
        # (multiple entities could share parents)
        for result in results:
            # groups can have a lot of members. converting the member list to a set keeps
            # the checking of all input entities to O(n) instead of O(n^2) and keeps the total
            # complexity of checking all input entities for all results to O(n^2) instead of O(n^3).
            # cast all members to lowercase for a case-insensitive membership check. our normalization
            # function gave use lowercase DNs
            member_set = set(member.lower() for member in result.get(ldap_constants.AD_ATTRIBUTE_MEMBER))
            for entity_dn in entity_dns:
                if entity_dn in member_set:
                    # get our input entity for the result dict
                    entity = entity_dns[entity_dn]
                    mapping_dict[entity].append(result)
        return mapping_dict

    def find_groups_for_group(self, group, attributes_to_lookup: List[str]=None):
        """ Find the groups that a group belongs to, look up attributes of theirs, and return information about them.

        :param group: The group to lookup group memberships for. This can either be an ADGroup or a string name of an
                      AD group. If it is a string, the group will be looked up first to get unique distinguished name
                      information about it unless it is a distinguished name.
        :param attributes_to_lookup: A list of string LDAP attributes to look up in addition to our basic attributes.
        :returns: A list of ADGroup objects representing the groups that this group belongs to.
        :raises: a DuplicateNameException if a group name is specified and more than one entry exists with the name.
        :raises: a InvalidLdapParameterException if the group name is not a string or ADGroup.
        """
        result_dict = self.find_groups_for_entities([group], attributes_to_lookup, self.find_group_by_name)
        return result_dict[group]

    def find_groups_for_groups(self, groups: List, attributes_to_lookup: List[str]=None):
        """ Find the groups that a list of groups belong to, look up attributes of theirs, and return information about
        them.

        :param groups: The groups to lookup group memberships for. This can be a list of either ADGroup objects or
                       string names of AD groups. If they are strings, the groups will be looked up first to get unique
                       distinguished name information about them unless they are distinguished names.
        :param attributes_to_lookup: A list of string LDAP attributes to look up in addition to our basic attributes.
        :returns: A dictionary mapping groups to lists of ADGroup objects representing the groups that they belong to.
        :raises: a DuplicateNameException if a group name is specified and more than one entry exists with the name.
        :raises: a InvalidLdapParameterException if any groups are not a string or ADGroup.
        """
        return self.find_groups_for_entities(groups, attributes_to_lookup, self.find_group_by_name)

    def find_groups_for_user(self, user, attributes_to_lookup: List[str]=None):
        """ Find the groups that a user belongs to, look up attributes of theirs, and return information about them.

        :param user: The user to lookup group memberships for. This can either be an ADUser or a string name of an
                     AD user. If it is a string, the user will be looked up first to get unique distinguished name
                     information about it unless it is a distinguished name.
        :param attributes_to_lookup: A list of string LDAP attributes to look up in addition to our basic attributes.
        :returns: A list of ADGroup objects representing the groups that this user belongs to.
        :raises: a DuplicateNameException if a user name is specified and more than one entry exists with the name.
        :raises: a InvalidLdapParameterException if the yser name is not a string or ADUser.
        """
        result_dict = self.find_groups_for_entities([user], attributes_to_lookup, self.find_user_by_name)
        return result_dict[user]

    def find_groups_for_users(self, users: List, attributes_to_lookup: List[str]=None):
        """ Find the groups that a list of users belong to, look up attributes of theirs, and return information about
        them.

        :param users: The users to lookup group memberships for. This can be a list of either ADUser objects or
                      string names of AD users. If they are strings, the users will be looked up first to get unique
                      distinguished name information about them unless they are distinguished names.
        :param attributes_to_lookup: A list of string LDAP attributes to look up in addition to our basic attributes.
        :returns: A dictionary mapping users to lists of ADGroup objects representing the groups that they belong to.
        :raises: a DuplicateNameException if a user name is specified and more than one entry exists with the name.
        :raises: a InvalidLdapParameterException if any users are not a string or ADUser.
        """
        return self.find_groups_for_entities(users, attributes_to_lookup, self.find_user_by_name)

    # FUNCTIONS FOR MODIFYING MEMBERSHIPS

    def _something_members_to_or_from_groups(self, members: List, groups_to_modify: List,
                                             member_lookup_fn: callable, stop_and_rollback_on_error: bool,
                                             adding: bool):
        """ Either add or remove members to/from groups. Members may be users or groups or string distinguished names.
        If there are any failures adding/removing for a group, and stop_and_rollback_on_error is True, we will attempt
        to undo the changes that have been done. If rollback fails, we will raise an exception. If it succeeds, we still
        raise an exception, but of a different type.

        If stop_and_rollback_on_error is False, we ignore failures and keep going.
        We return a list of successfully modified groups at the end.
        """
        # figure out which function to use
        member_modify_fn = self.ldap_connection.extend.microsoft.remove_members_from_groups
        verb = 'remove'
        if adding:
            member_modify_fn = self.ldap_connection.extend.microsoft.add_members_to_groups
            verb = 'add'

        # normalize our inputs to get lists of lowercase dns
        normalized_member_dns = ldap_utils.normalize_entities_to_entity_dns(members,
                                                                            member_lookup_fn)
        member_dn_list = list(normalized_member_dns.keys())
        normalized_target_group_dns = ldap_utils.normalize_entities_to_entity_dns(groups_to_modify,
                                                                                  self.find_group_by_name)
        target_group_list = list(normalized_target_group_dns.keys())

        # track successful groups in case we need to rollback, and rollback status
        successful_groups = []
        failing_group = None
        for group_dn in target_group_list:
            logger.info('Attempting to %s the following members to/from group %s : %s', verb, member_dn_list,
                        group_dn)
            # by setting fix to True, we ignore members already in the group and make this idempotent
            res = member_modify_fn(member_dn_list, [group_dn], fix=True)
            # the function returns a boolean on whether it succeeded or not
            if not res:
                failing_group = group_dn
                logger.error('Failure occurred attempting to %s members to/from group %s', verb, group_dn)
                if stop_and_rollback_on_error:
                    if not successful_groups:
                        logger.info('No successful groups to rollback')
                        break
                    logger.warning('Rolling back member %s operations to groups %s', verb, successful_groups)
                    # reverse the operation
                    new_op = not adding
                    try:
                        # don't attempt to rollback the rollback
                        self._something_members_to_or_from_groups(member_dn_list, successful_groups, member_lookup_fn,
                                                                  stop_and_rollback_on_error=False, adding=new_op)
                    except MembershipModificationException:
                        logger.error('Failed to completely rollback changes after failure. '
                                     'Halting and raising exception')
                        raise MembershipModificationRollbackException('Failed to modify group with distinguished name {} and rollback of '
                                                                      'successful groups with distinguished names {} failed. Please check '
                                                                      'the current state of those groups in AD before proceeding.'
                                                                      .format(group_dn, successful_groups))
                    logger.info('Successfully rolled back changes to %s', successful_groups)
                    break
                else:
                    logger.warning('Continuing despite failure')
            else:
                successful_groups.append(group_dn)

        # convert our successful groups into the input format before returning them
        original_groups_that_succeeded = [normalized_target_group_dns[group] for group in successful_groups]
        # raise an exception if we failed at all and wanted to stop and rollback
        # if a rollback was requested, then we only reach this point if it succeeds (a rollback exception is raised
        # if it fails), so raise an error declaring what failed and the fact that we rolled back
        if failing_group is not None and stop_and_rollback_on_error:
            raise MembershipModificationException('Failed to successfully update group {} to {} member. Rolled '
                                                  'back changes to other groups.'.format(failing_group, verb))
        return original_groups_that_succeeded

    def add_groups_to_groups(self, groups_to_add: List, groups_to_add_them_to: List,
                             stop_and_rollback_on_error: bool=True):
        return self._something_members_to_or_from_groups(groups_to_add, groups_to_add_them_to, self.find_group_by_name,
                                                         stop_and_rollback_on_error, adding=True)

    def add_users_to_groups(self, users_to_add: List, groups_to_add_them_to: List,
                            stop_and_rollback_on_error: bool=True):
        return self._something_members_to_or_from_groups(users_to_add, groups_to_add_them_to, self.find_user_by_name,
                                                         stop_and_rollback_on_error, adding=True)

    def remove_groups_from_groups(self, groups_to_remove: List, groups_to_remove_them_from: List,
                                  stop_and_rollback_on_error: bool=True):
        return self._something_members_to_or_from_groups(groups_to_remove, groups_to_remove_them_from,
                                                         self.find_group_by_name, stop_and_rollback_on_error,
                                                         adding=False)

    def remove_users_from_groups(self, users_to_remove: List, groups_to_remove_them_from: List,
                                 stop_and_rollback_on_error: bool=True):
        return self._something_members_to_or_from_groups(users_to_remove, groups_to_remove_them_from,
                                                         self.find_user_by_name, stop_and_rollback_on_error,
                                                         adding=False)

    def who_am_i(self):
        """ Return the authorization identity as recognized by the server.
        This can be helpful when a script is provided with an identity in one form that is used to start a session
        (e.g. a distinguished name, or a pre-populated kerberos cache) and then it wants to determine its identity
        that the server actually sees.
        This just calls the LDAP connection function, as it's suitable for AD as well.
        """
        return self.ldap_connection.extend.standard.who_am_i()

    def __repr__(self):
        conn_repr = self.ldap_connection.__repr__()
        domain_repr = self.domain.__repr__()
        return 'ADSession(ldap_connection={}, domain={})'.format(conn_repr, domain_repr)

    def __str__(self):
        return self.__repr__()



import copy

CERTIFICATE_AUTHORITY_OBJECT_CLASS = 'certificationAuthority'
TRUSTED_DOMAIN_OBJECT_CLASS = 'TrustedDomain'
COMPUTER_OBJECT_CLASS = 'computer'
GROUP_OBJECT_CLASS = 'group'
POSIX_GROUP_OBJECT_CLASS = 'posixGroup'
POSIX_USER_OBJECT_CLASS = 'posixAccount'
USER_OBJECT_CLASS = 'user'
TOP_OBJECT_CLASS = 'top'
# computers also have the user object class because they can act as users to operate
# within the domain, be a part of groups, etc.
OBJECT_CLASSES_FOR_COMPUTER = [COMPUTER_OBJECT_CLASS, USER_OBJECT_CLASS, TOP_OBJECT_CLASS]

# computers have an account control that determines things like whether they're trusted
# for
WORKSTATION_TRUST_ACCOUNT = 4096
DONT_EXPIRE_PASSWORD = 65536
COMPUTER_ACCESS_CONTROL_VAL = WORKSTATION_TRUST_ACCOUNT + DONT_EXPIRE_PASSWORD

# used to get all attributes that aren't derived from other attributes at the time of request.
# virtual attributes are only populated if specifically requested
AD_ATTRIBUTE_GET_ALL_NON_VIRTUAL_ATTRS = '*'

# keys for common active directory attributes
AD_ATTRIBUTE_SAMACCOUNT_NAME = 'sAMAccountName'
AD_ATTRIBUTE_SECURITY_DESCRIPTOR = 'ntSecurityDescriptor'
AD_ATTRIBUTE_COMMON_NAME = 'cn'
AD_ATTRIBUTE_OBJECT_CLASS = 'objectClass'

# keys for general user, group, and computer attributes
AD_ATTRIBUTE_USER_ACCOUNT_CONTROL = 'userAccountControl'
AD_ATTRIBUTE_SERVICE_PRINCIPAL_NAMES = 'servicePrincipalName'
AD_ATTRIBUTE_PASSWORD = 'unicodePwd'
# memberOf is a virtual attribute on users and groups, listing the DNs of groups that the record
# belongs to. it's less efficient to query because it's constructed on-demand in a lot of scenarios,
# and so we don't query for it by default, but use it if a user does query for it
AD_ATTRIBUTE_MEMBER_OF = 'memberOf'
# member is an attribute on groups, listing the DNs of users or groups that belong to it.
# it is part of the group record, so it's more efficient to query for and filter on
AD_ATTRIBUTE_MEMBER = 'member'
# posix attributes
AD_ATTRIBUTE_UID_NUMBER = 'uidNumber'  # posix user uid (uid is user name for ldap, uidNumber is uid)
AD_ATTRIBUTE_GID_NUMBER = 'gidNumber'  # posix group gid, or primary gid for user
AD_ATTRIBUTE_UNIX_HOME_DIR = 'unixHomeDirectory'  # homedir for a posix user
AD_ATTRIBUTE_UNIX_LOGIN_SHELL = 'loginShell'  # the login shell a user uses, e.g. /bin/bash, /bin/zsh

# keys for attributes that are relatively computer-specific
AD_ATTRIBUTE_ENCRYPTION_TYPES = 'msDS-SupportedEncryptionTypes'
AD_ATTRIBUTE_KVNO = 'msDS-KeyVersionNumber'
AD_ATTRIBUTE_DNS_HOST_NAME = 'dNSHostName'
AD_ATTRIBUTE_ADDITIONAL_DNS_HOST_NAME = 'msDS-AdditionalDnsHostName'

# keys for attributes in certificate authrities
AD_ATTRIBUTE_CA_CERT = 'caCertificate'

# keys for domains and trusted domains
AD_DOMAIN_FUNCTIONAL_LEVEL = 'domainFunctionality'
AD_DOMAIN_SUPPORTED_SASL_MECHANISMS = 'supportedSASLMechanisms'
AD_DOMAIN_TIME = 'currentTime'
AD_SCHEMA_VERSION = 'objectVersion'
AD_TRUSTED_DOMAIN_NAME = 'trustPartner'  # the netbios name of the trusted domain
AD_TRUST_TYPE = 'trustType'  # indicates windows or MIT
AD_TRUST_DIRECTION = 'trustDirection'  # disabled, incoming, outgoing, bidirectional
# TODO: implement decoding of more detailed trust info
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/e9a2d23c-c31e-4a6f-88a0-6646fdb51a3c
AD_TRUST_ATTRIBUTES = 'trustAttributes'


# From windows AD docs
AD_USERNAME_RESTRICTED_CHARS = {'[', ']', ':', ';', '|', '=', '+', '*', '?', '<', '>', '/', '\\',
                                '"', ','}
# max length for a normal sAMAccountName is 20 characters, including the '$' at the end
SAM_ACCOUNT_NAME_LENGTH = 20
# if NTLM needs to be supported or any legacy clients using UNC paths need to be supported,
# then sAMAccountName must be 16 characters or less, so the computer name must be 15 characters
# or less to allow for the trailing $
LEGACY_SAM_ACCOUNT_NAME_LENGTH_LIMIT = 16

# Active Directory standards for adding computers to a domain, as well as attribute and location
# related constants for domain-scope properties like dns servers and CAs
DEFAULT_COMPUTER_SERVICES = ['HOST']
DEFAULT_COMPUTER_LOCATION = 'CN=Computers'
DNS_SERVICE_FILTER = 'DNS/*'
DOMAIN_CONTROLLER_SCHEMA_VERSION_SEARCH_CONTAINER = 'CN=schema,CN=Configuration'
DOMAIN_WIDE_CONFIGURATIONS_CONTAINER = 'CN=Configuration'

# when checking if something simply exists, or getting everything at a level/subtree,
# we use this filter
FIND_ANYTHING_FILTER = '(objectClass=*)'
FIND_GROUP_FILTER = '(objectClass=Group)'
FIND_USER_FILTER = '(objectClass=User)'

# miscellaneous values we need
UNKNOWN_USER_POSIX_UID = -1
UNKNOWN_GROUP_POSIX_GID = -1


# the parent of all ADObjects, defined here to avoid risk of circular imports

class ADObject:

    def __init__(self, dn: str, attributes: dict, domain):
        self.distinguished_name = dn
        self.domain = domain
        self.all_attributes = attributes if attributes else {}
        # used for __repr__
        self.class_name = 'ADObject'

    def get(self, attribute_name: str, unpack_one_item_lists=False):
        """ Get an attribute about the group that isn't explicitly tracked as a member """
        val = self.all_attributes.get(attribute_name)
        # there's a lot of 1-item lists from the ldap3 library
        if isinstance(val, list) and len(val) == 1 and unpack_one_item_lists:
            return copy.deepcopy(val[0])
        return copy.deepcopy(val)

    def __repr__(self):
        attrs = self.all_attributes.__repr__() if self.all_attributes else 'None'
        domain = self.domain.__repr__()
        return ('{type}(dn={dn}, attributes={attrs}, domain={domain})'
                .format(type=self.class_name, dn=self.distinguished_name, attrs=attrs, domain=domain))

    def __str__(self):
        return self.__repr__()



import copy
import json
import pytz
import socket

from ms_active_directory import logging_utils

from datetime import datetime, timedelta, timezone
from ldap3 import (
    ANONYMOUS,
    BASE,
    Connection,
    FIRST,
    KERBEROS,
    NTLM,
    SAFE_RESTARTABLE,
    SASL,
    Server,
    ServerPool,
    SIMPLE,
    Tls,
)
from ssl import (
    OP_NO_SSLv2,
    OP_NO_SSLv3,
    OP_NO_TLSv1,
    OP_NO_TLSv1_1,
    CERT_NONE,
    CERT_REQUIRED,
)
from typing import List

# local imports come after imports from other libraries
from ms_active_directory.core.ad_session import ADSession
from ms_active_directory.environment.constants import ADFunctionalLevel
from ms_active_directory.environment.discovery.discovery_utils import discover_kdc_domain_controllers_in_domain, discover_ldap_domain_controllers_in_domain
from ms_active_directory.environment.format_utils import format_computer_name_for_authentication
from ms_active_directory.environment.kerberos.kerberos_constants import DEFAULT_KRB5_KEYTAB_FILE_LOCATION
from ms_active_directory.environment.ldap.ldap_constants import (
    AD_ATTRIBUTE_GET_ALL_NON_VIRTUAL_ATTRS,
    AD_DOMAIN_FUNCTIONAL_LEVEL,
    AD_DOMAIN_SUPPORTED_SASL_MECHANISMS,
    AD_DOMAIN_TIME,
    FIND_ANYTHING_FILTER,
)
from ms_active_directory.environment.ldap.ldap_format_utils import process_ldap3_conn_return_value
from ms_active_directory.environment.security.security_config_constants import ADEncryptionType
from ms_active_directory.exceptions import (
    DomainConnectException,
    DomainSearchException,
    InvalidDomainParameterException,
)


logger = logging_utils.get_logger()


def join_ad_domain(domain_dns_name: str, admin_user: str, admin_password: str, authentication_mechanism: str=SIMPLE,
                   ad_site: str=None, computer_name: str=None, computer_location: str=None, computer_password: str=None,
                   computer_encryption_types: List[ADEncryptionType]=None, computer_hostnames: List[str]=None,
                   computer_services: List[str]=None, supports_legacy_behavior: bool=False,
                   computer_key_file_path: str=DEFAULT_KRB5_KEYTAB_FILE_LOCATION, **additional_account_attributes):
    """ A super simple 'join a domain' function that requires minimal input - the domain dns name and admin credentials
    to use in the join process.
    Given those basic inputs, the domain's nearest controllers are automatically discovered and an account is made
    with strong security settings. The account's attributes follow AD naming conventions based on the computer's
    hostname by default.
    """
    domain = ADDomain(domain_dns_name, site=ad_site)
    return domain.join(admin_user, admin_password, authentication_mechanism, computer_name=computer_name,
                       computer_location=computer_location, computer_password=computer_password,
                       computer_hostnames=computer_hostnames, computer_services=computer_services,
                       computer_encryption_types=computer_encryption_types, supports_legacy_behavior=supports_legacy_behavior,
                       computer_key_file_path=computer_key_file_path, **additional_account_attributes)


def join_ad_domain_using_session(ad_session: ADSession, computer_name=None, computer_location=None,
                                 computer_password=None, computer_encryption_types=None, computer_hostnames=None,
                                 computer_services=None, supports_legacy_behavior=False,
                                 computer_key_file_path=DEFAULT_KRB5_KEYTAB_FILE_LOCATION,
                                 **additional_account_attributes):
    """ A fairly simple 'join a domain' function that requires minimal input - an AD session and admin credentials
    to use in the join process.
    Given those basic inputs, the domain's nearest controllers are automatically discovered and an account is made
    with strong security settings. The account's attributes follow AD naming conventions based on the computer's
    hostname by default.
    By providing an AD session, one can build a connection to the domain however they so choose and then use it to
    join this computer, so you don't even need to necessarily use user credentials.
    """
    # for joining a domain, default to using the local machine's hostname as a computer name
    if computer_name is None:
        computer_name = socket.gethostname()
        # if there's dots (e.g. our computer is server1.com) just take the first piece
        if '.' in computer_name:
            computer_name = computer_name.split('.')[0]
        logger.info('Using computer hostname (or its first component after splitting on dots) as computer name to '
                    'join domain: {}'.format(computer_name))
    logger.info('Attempting to join computer to domain %s with name %s', ad_session.get_domain_dns_name(),
                computer_name)
    computer = ad_session.create_computer(computer_name, computer_location=computer_location,
                                          computer_password=computer_password,
                                          encryption_types=computer_encryption_types, hostnames=computer_hostnames,
                                          services=computer_services, supports_legacy_behavior=supports_legacy_behavior,
                                          **additional_account_attributes)
    if computer_key_file_path is not None:
        computer.write_full_keytab_file_for_computer(computer_key_file_path)
    logger.info('Successfully joined computer to domain %s with name %s', ad_session.get_domain_dns_name(),
                computer_name)
    return computer


class ADDomain:

    def __init__(self, domain: str, site: str = None,
                 ldap_servers_or_uris: List = None,
                 kerberos_uris: List[str] = None,
                 encrypt_connections: bool = True,
                 ca_certificates_file_path: str = None,
                 discover_ldap_servers: bool = True,
                 discover_kerberos_servers: bool = True,
                 dns_nameservers: List[str] = None,
                 source_ip: str = None):
        """ Initialize an interface for defining an AD domain and interacting with it.
        :param domain: The DNS name of the Active Directory domain that this object represents.
        :param site: The Active Directory site to operate within. This is only relevant if LDAP or
                     kerberos servers are discovered in DNS, as there's site-specific records.
                     If set, only hosts within the specified site will be used.
        :param ldap_servers_or_uris: A list of either Server objects from the ldap3 library, or
                                     string LDAP uris. If specified, they will be used to establish
                                     sessions with the domain.
        :param kerberos_uris: A list of string kerberos server uris. These can be IPs (and the default
                              kerberos port of 88 will be used) or IP:port combinations.
        :param encrypt_connections: Whether or not LDAP connections with the domain will be secured
                                    using TLS. This must be True for join functionality to work,
                                    as passwords can only be set over secure connections.
                                    If not specified, defaults to True. If LDAP server objects are
                                    provided with ssl enabled or ldaps:// uris are provided, then
                                    connections to those servers will be encrypted because of the
                                    inherent behavior of such configurations.
        :param ca_certificates_file_path: A path to CA certificates to be used to establish trust
                                          with LDAP servers when securing connections. If not
                                          specified, then TLS will not check the peer certificate.
                                          If LDAP server objects are specified, then their TLS
                                          settings will be used rather than anything set in this
                                          variable. It is only used when discovering servers or
                                          using string URIs, so Server objects can be used if
                                          different CAs sign different servers' certificates
                                          due to regional CAs or something similar.
                                          If not specified, defaults to None.
        :param discover_ldap_servers: If true, and LDAP servers/uris are not specified, then LDAP
                                      servers for the domain will be discovered in DNS.
                                      If not specified, defaults to True.
        :param discover_kerberos_servers: If true, and kerberos uris are not specified, then kerberos
                                          servers for the domain will be discovered in DNS.
                                          If not specified, defaults to True.
        :param dns_nameservers: A list of strings indicating the IP addresses of DNS servers to use
                                when discovering servers for the domain. These may be IPv4 or IPv6
                                addresses.
                                If not specified, defaults to what's configured in /etc/resolv.conf on
                                POSIX systems, and extracting nameservers from registry keys on windows.
                                Defaults to None.
        :param source_ip: A source IP address to use for both DNS and LDAP connections established for
                          this domain. If not specified, defaults to automatic assignment of IP using
                          underlying system networking.
                          Defaults to None.
        """
        self.domain = domain.lower()  # cast to lowercase
        self.site = site.lower() if site else None
        self.encrypt_connections = encrypt_connections
        self.ca_certificates_file_path = ca_certificates_file_path
        self.ldap_servers = []
        self.ldap_uris = []
        self.kerberos_uris = []
        self.dns_nameservers = dns_nameservers
        self.source_ip = source_ip
        # discover ldap servers and kerberos servers if we weren't provided any and weren't told not to
        if not ldap_servers_or_uris and discover_ldap_servers:
            ldap_servers_or_uris = discover_ldap_domain_controllers_in_domain(self.domain, site=self.site,
                                                                              dns_nameservers=self.dns_nameservers,
                                                                              source_ip=self.source_ip,
                                                                              secure=self.encrypt_connections)
        # discover kerberos servers if we weren't provided any and weren't told not to
        if not kerberos_uris and discover_kerberos_servers:
            kerberos_uris = discover_kdc_domain_controllers_in_domain(self.domain, site=self.site,
                                                                      dns_nameservers=self.dns_nameservers,
                                                                      source_ip=self.source_ip)

        # handle the fact that user-provided ldap servers could be servers or strings
        if ldap_servers_or_uris:
            self.set_ldap_servers_or_uris(ldap_servers_or_uris)

        if kerberos_uris:
            self.set_kerberos_uris(kerberos_uris)

    def get_domain_dns_name(self):
        return self.domain

    def get_ldap_servers(self):
        return copy.deepcopy(self.ldap_servers)

    def get_ldap_uris(self):
        return copy.deepcopy(self.ldap_uris)

    def get_kerberos_uris(self):
        return copy.deepcopy(self.kerberos_uris)

    def set_ldap_servers_or_uris(self, ldap_servers_or_uris: List):
        """ Set our list of LDAP servers or LDAP URIs. The list provided can be a list of
        Server objects, URIs, or a mixture.
        """
        ldap_uris = []
        ldap_server_objs = []
        # users can specify Server objects if they want a custom Tls setting for
        # each one, but if they provide strings then we just make our own for each
        for serv in ldap_servers_or_uris:
            if isinstance(serv, str):
                # pass through whatever CA certs file we got, and disable all TLS below
                # 1.2 by default
                tls_setting = None
                if self.encrypt_connections:
                    # only check peer certs if we have CAs
                    checking = CERT_REQUIRED if self.ca_certificates_file_path else CERT_NONE
                    tls_setting = Tls(ca_certs_file=self.ca_certificates_file_path,
                                      ssl_options=[OP_NO_SSLv2, OP_NO_SSLv3, OP_NO_TLSv1,
                                                   OP_NO_TLSv1_1],
                                      validate=checking)
                ldap_server_objs.append(Server(serv, tls=tls_setting))
                ldap_uris.append(serv)
            elif isinstance(serv, Server):
                # extract the uri for list of the server uris we're using
                ldap_server_objs.append(serv)
                ldap_uris.append(serv.name)
            else:
                raise InvalidDomainParameterException('Invalid type for element of ldap server list, {}; '
                                                      'elements must be strings or Server objects'.format(type(serv)))
        self.ldap_servers = ldap_server_objs
        self.ldap_uris = ldap_uris

    def set_kerberos_uris(self, kerberos_uris: List):
        """ Sets our kerberos server uris """
        for serv in kerberos_uris:
            if not isinstance(serv, str):
                raise InvalidDomainParameterException('Invalid type for element of kerberos server list, {}; '
                                                      'elements must be strings'.format(type(serv)))
        self.kerberos_uris = kerberos_uris

    def is_close_in_time_to_localhost(self, ldap_connection: Connection=None, allowed_drift_seconds: int=None):
        """ Check if we're close in time to the domain.
        This is primarily useful for kerberos and TLS negotiation health.
        Optionally, an existing connection can be used. If one is not specified, an anonymous LDAP
        connection will be created and used.
        :param ldap_connection: An ldap3 connection to the domain, optional.
        :param allowed_drift_seconds: The number of seconds considered "close", defaults to 5 minutes.
                                      5 minutes is the standard allowable drift for kerberos.
        :return: A boolean indicating whether we're within allowed_drift_seconds seconds of the domain time.
        """
        if allowed_drift_seconds is None:
            allowed_drift_seconds = 300
        domain_time = self.find_current_time(ldap_connection)
        local_time = datetime.now(tz=timezone.utc)
        diff = domain_time - local_time
        if local_time > domain_time:
            diff = local_time - domain_time
        return diff < timedelta(seconds=allowed_drift_seconds)

    def find_current_time(self, ldap_connection: Connection=None):
        """ Find the current time for this domain. This is useful for detecting drift that can cause
        Kerberos and TLS issues.
        Optionally, an existing connection can be used. If one is not specified, an anonymous LDAP
        connection will be created and used.
        :param ldap_connection: An ldap3 connection to the domain, optional.
        :return: A datetime object representing the time.
        """
        if ldap_connection is None:
            logger.info('Creating a new anonymous connection to read domain time')
            # we just need an anonymous session to read this information
            ldap_connection = self.create_session_as_user().get_ldap_connection()

        res = ldap_connection.search(search_base='', search_filter=FIND_ANYTHING_FILTER,
                                     # querying for time as an attribute explicitly doesn't work since it's not a real
                                     # ldap attribute in any rfc
                                     search_scope=BASE, attributes=[AD_ATTRIBUTE_GET_ALL_NON_VIRTUAL_ATTRS],
                                     size_limit=1)
        success, _, response, _ = process_ldap3_conn_return_value(ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search the domain to query the current time.')
        base_attrs = response[0]['attributes']
        # time comes back as a 1-item list in the format ["20210809080919.0Z"]
        # that's a date string yyyyMMddHHmmss.0Z
        ad_time = base_attrs.get(AD_DOMAIN_TIME)[0]
        useful_time = datetime.strptime(ad_time, '%Y%m%d%H%M%S.0Z')
        return pytz.utc.localize(useful_time)

    def find_functional_level(self, ldap_connection: Connection=None):
        """ Find the functional level for this domain.
        Optionally, an existing connection can be used. If one is not specified, an anonymous LDAP
        connection will be created and used.
        :param ldap_connection: An ldap3 connection to the domain, optional.
        :return: An ADVersion enum indicating the functional level.
        """
        if ldap_connection is None:
            logger.info('Creating a new anonymous connection to read the domain functional level')
            # we just need an anonymous session to read this information
            ldap_connection = self.create_session_as_user().get_ldap_connection()

        res = ldap_connection.search(search_base='', search_filter=FIND_ANYTHING_FILTER,
                                     search_scope=BASE, attributes=[AD_ATTRIBUTE_GET_ALL_NON_VIRTUAL_ATTRS],
                                     size_limit=1)
        success, _, response, _ = process_ldap3_conn_return_value(ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed search the domain for its functional level')
        base_attrs = response[0]['attributes']
        # this is a single-item list of a string of the level number, like ["7"]
        level_str = base_attrs.get(AD_DOMAIN_FUNCTIONAL_LEVEL)[0]
        return ADFunctionalLevel.get_functional_level_from_value(int(level_str))

    def find_supported_sasl_mechanisms(self, ldap_connection: Connection=None):
        """ Find the supported SASL mechanisms for this domain.
        Optionally, an existing connection can be used. If one is not specified, an anonymous LDAP
        connection will be created and used.
        :param ldap_connection: An ldap3 connection to the domain, optional.
        :return: A list of strings indicating the supported SASL mechanisms for the domain.
                 ex: ['GSSAPI', 'GSS-SPNEGO', 'EXTERNAL']
        """
        if ldap_connection is None:
            logger.info('Creating a new anonymous connection to read domain supported SASL mechanisms')
            # we just need an anonymous session to read this information
            ldap_connection = self.create_session_as_user().get_ldap_connection()

        res = ldap_connection.search(search_base='', search_filter=FIND_ANYTHING_FILTER,
                                     search_scope=BASE, attributes=[AD_DOMAIN_SUPPORTED_SASL_MECHANISMS],
                                     size_limit=1)
        success, _, response, _ = process_ldap3_conn_return_value(ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search the domain for supported SASL mechanisms.')
        base_attrs = response[0]['attributes']
        return base_attrs.get(AD_DOMAIN_SUPPORTED_SASL_MECHANISMS, [])

    def refresh_ldap_server_discovery(self):
        """ Re-discover LDAP servers in DNS for the domain and redo the sorting by RTT.
        This can update our list of LDAP servers for future connections, allowing faster servers to be
        moved up in priority, unavailable servers to be removed from the list, and previously unavailable
        servers to be added.
        """
        ldap_uris = discover_ldap_domain_controllers_in_domain(self.domain, site=self.site,
                                                               secure=self.encrypt_connections,
                                                               dns_nameservers=self.dns_nameservers,
                                                               source_ip=self.source_ip)
        self.set_ldap_servers_or_uris(ldap_uris)

    def refresh_kerberos_server_discovery(self):
        """ Re-discover Kerberos servers in DNS for the domain and redo the sorting by RTT.
        This can update our list of KDCs for future use by callers, allowing faster servers to be
        moved up in priority, unavailable servers to be removed from the list, and previously unavailable
        servers to be added.
        """
        kerberos_uris = discover_kdc_domain_controllers_in_domain(self.domain, site=self.site,
                                                                  dns_nameservers=self.dns_nameservers,
                                                                  source_ip=self.source_ip)
        self.set_kerberos_uris(kerberos_uris)

    def _create_connection(self, user, password, authentication_mechanism, **kwargs):
        """ Internal helper for creating sessions regardless of whether they're for users or computers """
        if len(self.ldap_servers) == 0:
            raise DomainConnectException('Cannot create a session with the AD domain, as there are no LDAP servers '
                                         'known for the domain.')
        # our servers were either user specified (in which case it's a list of ordered preferences) or
        # were discovered automatically (in which case they're ordered by RTT), so use the FIRST strategy
        # to either contact the first preferred server or the fastest/closest server
        server_pool = ServerPool(servers=self.ldap_servers, pool_strategy=FIRST)
        # if no client strategy is specified, default to restartable. AD tends to close idle connections;
        # also if a user specified the LDAP servers, they may have used a hostname that has many servers
        # behind it (like just the domain name), which can cause connections to break if they're using TLS
        # an experience a dns failover on a synchronous connection. Restartable connections avoid these issues.
        # Use safe restartable in case the caller uses this in a multi-threaded application.
        if not kwargs.get('client_strategy'):
            conn = Connection(server_pool, user=user, password=password, authentication=authentication_mechanism,
                              client_strategy=SAFE_RESTARTABLE, source_address=self.source_ip,
                              **kwargs)
        else:
            conn = Connection(server_pool, user=user, password=password, authentication=authentication_mechanism,
                              source_address=self.source_ip, **kwargs)

        conn.open()
        logger.debug('Opened connection to AD domain %s: %s', self.domain, conn)
        if self.encrypt_connections:
            # if we're using LDAPS, don't StartTLS
            if not conn.server.ssl:
                tls_started = conn.start_tls()
                if not tls_started:
                    raise DomainConnectException('Unable to StartTLS on connection to domain. Please check the '
                                                 'server(s) to ensure that they have properly configured certificates.')
            logger.debug('Successfully secured connection to AD domain %s', self.domain)
        bind_resp = conn.bind()
        bound, result, _, _ = process_ldap3_conn_return_value(conn, bind_resp)
        if not bound:
            raise DomainConnectException('Failed to bind connection to {} - please check the credentials and '
                                         'authentication mechanism in use. LDAP result: {}'
                                         .format(conn.server.name, result))
        logger.debug('Successfully bound connection to AD domain %s to establish session', self.domain)
        return conn

    def create_ldap_connection_as_user(self, user: str=None, password: str=None, authentication_mechanism: str=None,
                                       **kwargs):
        """ Create an LDAP connection with AD domain authenticated as the specified user. """
        logger.info('Establishing connection with AD domain %s using LDAP authentication mechanism %s and user %s',
                    self.domain, authentication_mechanism, user)
        return self._create_connection(user, password, authentication_mechanism, **kwargs)

    def create_session_as_user(self, user: str=None, password: str=None, authentication_mechanism: str=None,
                               **kwargs):
        """ Create a session with AD domain authenticated as the specified user. """
        logger.info('Establishing session with AD domain %s using LDAP authentication mechanism %s and user %s',
                    self.domain, authentication_mechanism, user)
        conn = self._create_connection(user, password, authentication_mechanism, **kwargs)
        session = ADSession(conn, self)
        return session

    def create_ldap_connection_as_computer(self, computer_name: str, computer_password: str=None, check_name_format: bool=True,
                                           authentication_mechanism: str=KERBEROS, **kwargs):
        """ Create an LDAP connection with AD domain authenticated as the specified computer. """
        logger.info('Establishing LDAP connection with AD domain %s using LDAP authentication mechanism %s and computer %s',
                    self.domain, authentication_mechanism, computer_name)
        # reject simple binds because computers can't use them for authentication
        if authentication_mechanism == SIMPLE or authentication_mechanism == ANONYMOUS:
            raise InvalidDomainParameterException('Computers must use a form of SASL or NTLM for authenticating LDAP '
                                                  'communication with and AD domain.')
        # when using EXTERNAL authentication (certificate-based) there might be some weird names, so let power users
        # skip our helpful name formatting/validation logic in case it causes issues
        formatted_name = computer_name
        if check_name_format:
            formatted_name = format_computer_name_for_authentication(computer_name, self.domain,
                                                                     authentication_mechanism)
        # ntlm isn't real SASL, but Kerberos, EXTERNAL, etc. are
        if authentication_mechanism != NTLM:
            kwargs['sasl_mechanism'] = authentication_mechanism
            authentication_mechanism = SASL
        return self._create_connection(formatted_name, computer_password, authentication_mechanism, **kwargs)

    def create_session_as_computer(self, computer_name: str, computer_password: str=None, check_name_format: bool=True,
                                   authentication_mechanism: str=KERBEROS, **kwargs):
        """ Create a session with AD domain authenticated as the specified computer. """
        logger.info('Establishing session with AD domain %s using LDAP authentication mechanism %s and computer %s',
                    self.domain, authentication_mechanism, computer_name)
        conn = self.create_ldap_connection_as_computer(computer_name, computer_password, check_name_format,
                                                       authentication_mechanism, **kwargs)
        session = ADSession(conn, self)
        return session

    def join(self, admin_username: str, admin_password: str, authentication_mechanism: str=SIMPLE,
             computer_name: str=None, computer_location: str=None, computer_password: str=None,
             computer_encryption_types: List[ADEncryptionType]=None, computer_hostnames: List[str]=None,
             computer_services: List[str]=None, supports_legacy_behavior: bool=False,
             computer_key_file_path: str=DEFAULT_KRB5_KEYTAB_FILE_LOCATION,
             **additional_account_attributes):
        """ A super simple 'join the domain' function that requires minimal input - just admin user credentials
        to use in the join process.
        Given those basic inputs, the domain's settings are used to establish a connection, and an account is made
        with strong security settings. The account's attributes follow AD naming conventions based on the computer's
        hostname by default.
        """
        ad_session = self.create_session_as_user(admin_username, admin_password, authentication_mechanism)
        return join_ad_domain_using_session(ad_session, computer_name=computer_name, computer_location=computer_location,
                                            computer_password=computer_password, computer_hostnames=computer_hostnames,
                                            computer_services=computer_services,
                                            computer_encryption_types=computer_encryption_types,
                                            supports_legacy_behavior=supports_legacy_behavior,
                                            computer_key_file_path=computer_key_file_path,
                                            **additional_account_attributes)

    def __repr__(self):
        result = 'ADDomain(domain={}'.format(self.domain)
        if self.site:
            result += ', site={}'.format(self.site)
        enc_connections = 'True' if self.encrypt_connections else 'False'
        result += ', encrypt_connections=' + enc_connections
        if self.ldap_servers:
            list_repr = ','.join(serv.__repr__() for serv in self.ldap_servers)
            result += ', ldap_servers_or_uris=[{}]'.format(list_repr)
        if self.kerberos_uris:
            list_repr = ','.join(serv.__repr__() for serv in self.kerberos_uris)
            result += ', kerberos_uris=[{}]'.format(list_repr)
        if self.ca_certificates_file_path:
            result += ', ca_certificates_file_path=' + self.ca_certificates_file_path
        if self.dns_nameservers:
            list_repr = ','.join(serv.__repr__() for serv in self.dns_nameservers)
            result += ', dns_nameservers=[{}]'.format(list_repr)
        if self.source_ip:
            result += ', source_ip=' + self.source_ip
        result += ')'

        return result

    def __str__(self):
        return self.__repr__()



from ms_active_directory import logging_utils

from ldap3 import Connection
from ldap3.core.exceptions import LDAPInvalidDnError
from ldap3.utils.dn import parse_dn
from typing import List

from ms_active_directory.environment.ldap.ldap_constants import (
    ADObject,
    AD_USERNAME_RESTRICTED_CHARS,
    SAM_ACCOUNT_NAME_LENGTH,
    LEGACY_SAM_ACCOUNT_NAME_LENGTH_LIMIT,
)
from ms_active_directory.exceptions import (
    InvalidDomainParameterException,
    InvalidLdapParameterException,
    ObjectNotFoundException,
)


logger = logging_utils.get_logger()


def is_dn(anything: str):
    """ Determine if a specified string is a distinguished name. """
    try:
        # our sessions all set check_names to the default value of True, so dns will be escaped
        # in search bases and such. therefore we escape them here.
        # if this is a check for filtering, then we'll be getting the unescaped value to figure out
        # how to best escape it, so definitely escape it to avoid errors from special characters
        parse_dn(anything, escape=True)
        return True
    except LDAPInvalidDnError:
        return False


def construct_default_hostnames_for_computer(computer_name: str, domain_dns_name: str):
    """ Construct the default hostnames for a computer in AD. The short hostname is the computer name capitalized,
    and the fqdn is lowercase of the computer name dot the domain.
    """
    return [computer_name.upper(), computer_name.lower() + '.' + domain_dns_name.lower()]


def construct_object_distinguished_name(object_name: str, object_location: str, domain: str):
    """
    Constructs the distinguished name of a computer, group, or user given the name, join location, and domain.
    """
    computer_part = 'CN=' + object_name
    domain_part = construct_ldap_base_dn_from_domain(domain)
    return ','.join([computer_part, object_location, domain_part])


def construct_domain_from_ldap_base_dn(domain: str):
    """
    Given a base DN, constructs the DNS name of the AD domain.
    """
    dn_split = parse_dn(domain)
    # parse dn takes "cn=demo,ou=Computers,dc=example,DC=com" and turns it into
    # [('cn', 'demo', ','), ('ou', 'Computers', ','), ('dc', 'example', ','), ('DC', 'com', '')]
    domain_pieces = [piece[1] for piece in dn_split if piece[0].upper() == 'DC']
    return '.'.join(domain_pieces)


def construct_ldap_base_dn_from_domain(domain: str):
    """
    Given a domain, constructs the base dn.
    """
    domain_split = domain.split('.')
    return ','.join(map(lambda x: 'DC=' + x, domain_split))


def construct_service_principal_names(services: List[str], hostnames: List[str]):
    """ Given a list of services and hostnames, construct the kerberos server principle names for them. """
    spns = []
    for serv in services:
        for hostname in hostnames:
            spns.append(serv + '/' + hostname)
    return spns


def escape_generic_filter_value(anything: str):
    """ Escape anything, so that it can be used in ldap queries without confusing the server.
    According to the LDAP spec, there's a set of common characters that need escaping:
    rfc4514 (https://tools.ietf.org/html/rfc4514).

    RFCs that define new LDAP attributes, as well different server types, may require
    additional characters be escaped. Additionally, not all characters need to be escaped.
    For example, many versions of AD do not require commas be escaped, but will be ok if
    they are. Please ensure you know what you're escaping before calling this.
    See escape_dn_for_filter for an example of an alternative escape function needed to
    escape a field with different properties.
    """
    if anything.isalnum():
        return anything

    def escape_char(char):
        """ Escape a single character."""
        if char in "*()\\/\0 \t\r\n+<>,\";":
            # LDAP query language is really forgiving about strange characters.
            # rfc2254 says the only characters to escape are "*{}\\\0". AD adds "/" to the
            # list, and OpenLDAP adds whitespace. Over-escaping is safe, so just do everything
            # every time.
            return "\\%02x" % ord(char)
        else:
            return char
    return "".join(escape_char(x) for x in anything)


def escape_dn_for_filter(anything: str):
    """Escape an LDAP distinguished name so that it can be used in filters without confusing the server.
    Distinguished names already have some special characters escaped or encoded, so we must use this
    function instead of the generic escape function, which would escape the existing escape sequences.

    In a filter, you use the format field=value.
    But distinguished names are in the form CN=x,OU=y,DC=z so those equal signs need to be escaped.
    But then the values x, y, and z can also have equal signs in them, and those will ALREADY be escaped
    differently from the ones following CN, OU, etc.
    That's why DNs need a different escaping in filters than everything else.
    """
    if isinstance(anything, int) or isinstance(anything, float):
        return anything

    if anything.isalnum():
        return anything

    def escape_char(char):
        """ Escape a single character."""
        if char in "()*":
            return "\\%02x" % ord(char)
        else:
            return char
    return "".join(escape_char(x) for x in anything)


def normalize_entities_to_entity_dns(entities: List, lookup_by_name_fn: callable):
    """ Given a list of entities that might be AD objects or strings, return a map of LDAP distinguished names
    for the entities.
    """
    # make a map of entity distinguished names to entities passed in. we'll use this when constructing
    # our return dictionary as well
    entity_dns = {}
    for entity in entities:
        if isinstance(entity, str):
            if is_dn(entity):
                # cast to lowercase for case-insensitive checks later
                entity_dn = entity.lower()
            elif lookup_by_name_fn is None:
                raise InvalidLdapParameterException('If any entities are strings and are not distinguished names, '
                                                    'a function must be provided to look up entities by name. '
                                                    'Context: {}'.format(entity))
            else:
                entity_obj = lookup_by_name_fn(entity)
                if entity_obj is None:
                    raise ObjectNotFoundException('No entity could be found with name {}'.format(entity))
                # cast to lowercase for case-insensitive checks later
                entity_dn = entity_obj.distinguished_name.lower()
        elif isinstance(entity, ADObject):
            # cast to lowercase for case-insensitive checks later
            entity_dn = entity.distinguished_name.lower()
        else:
            bad_type = type(entity)
            raise InvalidLdapParameterException('All entities must either be ADObject objects or strings. {} was '
                                                'of type {}'.format(entity, bad_type))
        entity_dns[entity_dn] = entity
    return entity_dns


def normalize_object_location_in_domain(location: str, domain_dns_name: str):
    """ There's two main formats we might see used for an object location - LDAP style and Windows Path style.
    For each style, they can be relative or fully qualified.

    LDAP Style looks like this:
    OU=Location
    or fully qualified:
    OU=Location,DC=example,DC=com

    Windows Path Style looks like this:
    computers/
    or fully qualified:
    example.com/computers

    This function tries to normalize everything to the LDAP style using the relative DN format, since the library
    is primarily LDAP based.
    Currently, windows path style will be rejected. This is because we'd need to make LDAP queries at each level
    to determine the proper naming attribute for the next level (e.g. CN or OU). We can do that, but we'd need a
    connection already established or credentials.
    """
    # TODO: make windows path style names work
    if not is_dn(location):
        raise NotImplementedError('Windows-style paths are not supported')
    return strip_domain_from_object_location(location, domain_dns_name)


def process_ldap3_conn_return_value(ldap_connection: Connection, return_value):
    """ Thread-safe ldap3 connections return a tuple containing a boolean about success,
    the result, the response, and the request. Non-thread-safe ldap3 connections just
    leave the other fields and return a boolean when performing search/add/etc. and
    leave it up to the caller to manage thread safety.

    This function processes the return value so that it can be used within this class
    without worrying about the return format.
    """
    if ldap_connection.strategy.thread_safe:
        success, result, response, req = return_value
    else:
        success = return_value
        result = ldap_connection.result
        response = ldap_connection.response
        req = ldap_connection.request
    return success, result, response, req


def remove_ad_search_refs(response):
    """ Many LDAP queries in Active Directory will include a number of generic search references
    to say 'maybe go look here for completeness'. This is especially common in setups where
    there's trusted domains or other domains in the same forest.

    But for many domain operations, we don't care about them. For example, when checking for other
    accounts with a sAMAccountName in a domain, we don't care about the possibility of accounts
    existing with that name in another domain in the forest because it's a domain-unique
    attribute.
    So this is a helper function to remove such references.

    :param response: A list of LDAP search responses.
    :return: A filtered list, with search references removed.
    """
    if not response:
        return []
    real_entities = []
    if response:
        real_entities = [entry for entry in response if entry.get('dn')]
    return real_entities


def strip_domain_from_object_location(location: str, domain_dns_name: str):
    """ Our object Location in a domain should be a relative distinguished name (RDN), but if someone specifies the full
    path, let's be forgiving.
    This is a normalizing function to convert to RDNs.
    So if a user specifies "OU=Location,DC=example,DC=com" this function will strip off "DC=example,DC=com"
    and leave the relative distinguished name "OU=Location" which is what we'll actually use.
    """
    if location is None:
        return location

    # cast everything to uppercase in order to avoid worrying about how a customer chose to type their DN.
    # place a comma in front of the domain RDN so that any stripping we do will strip the trailing comma
    domain_rdn_upper = ',' + construct_ldap_base_dn_from_domain(domain_dns_name).upper()
    location = location.upper()
    if location.endswith(domain_rdn_upper):
        trim_length = len(domain_rdn_upper)
        # trim the length of our domain RDN from the end
        location = location[:-trim_length]

    return location


def validate_and_normalize_computer_name(name: str, supports_legacy_behavior: bool):
    """ Computer common names are sAMAccountNames without the $ at the end. So check for allowable
    characters and length limits.
    """
    limit = LEGACY_SAM_ACCOUNT_NAME_LENGTH_LIMIT if supports_legacy_behavior else SAM_ACCOUNT_NAME_LENGTH
    # peel off the ending $ if present
    if name.endswith('$'):
        name = name[:-1]
    if len(name) > limit:
        raise InvalidDomainParameterException('Computer name length must be fewer than {} characters for computers '
                                              'that {} legacy behavior.'.format(limit, 'support' if supports_legacy_behavior else 'do not support'))
    for character in AD_USERNAME_RESTRICTED_CHARS:
        if character in name:
            raise InvalidDomainParameterException('AD computer names may not contain any of the following characters: '
                                                  '{}'.format(', '.join(AD_USERNAME_RESTRICTED_CHARS)))
    return name



""" Constants describing AD Domains that aren't specific to any protocol """
import enum


class ADVersion(enum.Enum):
    """ This enum maps AD version friendly names to schema versions """
    WINDOWS_UNRECOGNIZED = None
    WINDOWS_2000 = 13
    # Active Directory application mode shares a schema with 2003
    WINDOWS_2003_OR_ADAM = 30
    # Active Directory lightweight directory services shares a schema with 2003R2
    WINDOWS_2003R2_OR_AD_LDS = 31
    WINDOWS_2008 = 44
    WINDOWS_2008R2 = 47
    WINDOWS_2012 = 56
    WINDOWS_2012R2 = 69
    WINDOWS_2016 = 87
    WINDOWS_2019 = 88

    @classmethod
    def get_version_from_schema_number(cls, schema):
        for ver in ADVersion:
            if ver.value == schema:
                return ver
        return ADVersion.WINDOWS_UNRECOGNIZED


class ADFunctionalLevel(enum.Enum):
    """ This enum maps AD Domain functional level friendly names to values """
    WINDOWS_UNRECOGNIZED = None
    WINDOWS_2000 = 0
    WINDOWS_2003 = 2
    WINDOWS_2008 = 3
    WINDOWS_2008R2 = 4
    WINDOWS_2012 = 5
    WINDOWS_2012R2 = 6
    WINDOWS_2016 = 7

    @classmethod
    def get_functional_level_from_value(cls, val):
        for ver in ADFunctionalLevel:
            if ver.value == val:
                return ver
        return ADFunctionalLevel.WINDOWS_UNRECOGNIZED



import logging

_ad_logger = None


def configure_log_level(level: str):
    """ Set the log level of the AD logger
    :param level: The lowest log severity to be recorded.
    """
    get_logger().setLevel(level)


def disable_logging():
    """ Disable logging entirely for the AD logger """
    get_logger().propagate = False


def enable_logging():
    """ Enable logging for the AD logger """
    get_logger().propagate = True


def get_logger():
    """ Retrieve the AD logger for this package. If it has not been declared, then declare it. """
    global _ad_logger
    if _ad_logger is not None:
        return _ad_logger
    logger = logging.getLogger('ms_active_directory')
    # by default, only log info+
    logger.setLevel(logging.INFO)
    _ad_logger = logger
    return _ad_logger



#!/usr/bin/env python3

from setuptools import setup

version = '1.1.0'
author = 'Azaria Zornberg'
email = 'a.zornberg96@gmail.com'
license_str = 'MIT License'
url = 'https://github.com/zorn96/ms_active_directory/'
description = 'Python library for integrating with Microsoft Active Directory'
package_name = 'ms_active_directory'
package_folder = '.'

long_description = open('README.md', encoding='utf-8').read()
packages = ['ms_active_directory',
            'ms_active_directory.core',
            'ms_active_directory.environment',
            'ms_active_directory.environment.discovery',
            'ms_active_directory.environment.kerberos',
            'ms_active_directory.environment.ldap',
            'ms_active_directory.environment.security'
            ]


setup_kwargs = {
    'packages': packages,
    'package_dir': {'': package_folder},
}

requirements = ['dnspython>=2.1.0',
                'ldap3>=2.8.0',
                'pyasn1>=0.4.6',
                'pycryptodome>=3.9.0',
                'pytz',
                'six>=1.15.0',
                ]

setup(name=package_name,
      version=version,
      install_requires=requirements,
      license=license_str,
      author=author,
      author_email=email,
      description=description,
      long_description=long_description,
      long_description_content_type='text/markdown',
      keywords='python3 ldap microsoft windows active-directory kerberos ad',
      python_requires=">=3.5",
      url=url,
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'Intended Audience :: System Administrators',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: MacOS :: MacOS X',
                   'Operating System :: Microsoft :: Windows',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Topic :: Security',
                   'Topic :: Software Development :: Libraries :: Python Modules',
                   'Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP'],
      **setup_kwargs
      )



import copy

from ms_active_directory.environment.ldap.ldap_constants import (
    ADObject,
    AD_ATTRIBUTE_COMMON_NAME,
    AD_ATTRIBUTE_GID_NUMBER,
    AD_ATTRIBUTE_OBJECT_CLASS,
    AD_ATTRIBUTE_SAMACCOUNT_NAME,
    AD_ATTRIBUTE_UID_NUMBER,
    AD_ATTRIBUTE_UNIX_HOME_DIR,
    AD_ATTRIBUTE_UNIX_LOGIN_SHELL,
    UNKNOWN_GROUP_POSIX_GID,
    UNKNOWN_USER_POSIX_UID,
)
from ms_active_directory.environment.ldap.ldap_format_utils import (
    normalize_object_location_in_domain
)


class ADUser(ADObject):

    def __init__(self, dn: str, attributes: dict, domain):
        super().__init__(dn, attributes, domain)
        # used for __repr__
        self.class_name = 'ADUser'
        self.samaccount_name = attributes.get(AD_ATTRIBUTE_SAMACCOUNT_NAME)
        self.common_name = attributes.get(AD_ATTRIBUTE_COMMON_NAME)
        self.object_classes = attributes.get(AD_ATTRIBUTE_OBJECT_CLASS)

        self.name = self.samaccount_name
        # to get the location of a user, remove their common name from the front and the domain's
        # domain components from the end
        rdn_of_user = normalize_object_location_in_domain(dn, self.domain.get_domain_dns_name())
        user_piece = 'CN=' + self.common_name
        self.location = rdn_of_user[len(user_piece)+1:]


class ADPosixUser(ADUser):

    def __init__(self, dn: str, attributes: dict, domain):
        super().__init__(dn, attributes, domain)
        # used for __repr__
        self.class_name = 'ADPosixUser'
        # posix attributes aren't actually mandatory for posix objects in AD, so we're not guaranteed these
        # exist
        self.uid = UNKNOWN_USER_POSIX_UID
        # uidNumber will come back as a singleton list
        if AD_ATTRIBUTE_UID_NUMBER in attributes and attributes.get(AD_ATTRIBUTE_UID_NUMBER):
            self.uid = int(attributes.get(AD_ATTRIBUTE_UID_NUMBER)[0])

        self.gid = UNKNOWN_GROUP_POSIX_GID
        # gidNumber will come back as a singleton list
        if AD_ATTRIBUTE_GID_NUMBER in attributes and attributes.get(AD_ATTRIBUTE_GID_NUMBER):
            self.gid = int(attributes.get(AD_ATTRIBUTE_GID_NUMBER)[0])

        self.unix_home_directory = None
        if AD_ATTRIBUTE_UNIX_HOME_DIR in attributes and attributes.get(AD_ATTRIBUTE_UNIX_HOME_DIR):
            self.unix_home_directory = attributes.get(AD_ATTRIBUTE_UNIX_HOME_DIR)[0]

        self.login_shell = None
        if AD_ATTRIBUTE_UNIX_LOGIN_SHELL in attributes and attributes.get(AD_ATTRIBUTE_UNIX_LOGIN_SHELL):
            self.login_shell = attributes.get(AD_ATTRIBUTE_UNIX_LOGIN_SHELL)[0]


class ADGroup(ADObject):

    def __init__(self, dn: str, attributes: dict, domain):
        super().__init__(dn, attributes, domain)
        # used for __repr__
        self.class_name = 'ADGroup'
        self.samaccount_name = attributes.get(AD_ATTRIBUTE_SAMACCOUNT_NAME)
        self.common_name = attributes.get(AD_ATTRIBUTE_COMMON_NAME)
        self.object_classes = attributes.get(AD_ATTRIBUTE_OBJECT_CLASS)

        self.name = self.samaccount_name
        # to get the location of a user, remove their common name from the front and the domain's
        # domain components from the end
        rdn_of_user = normalize_object_location_in_domain(dn, self.domain.get_domain_dns_name())
        user_piece = 'CN=' + self.common_name
        self.location = rdn_of_user[len(user_piece)+1:]


class ADPosixGroup(ADGroup):

    def __init__(self, dn: str, attributes: dict, domain):
        super().__init__(dn, attributes, domain)
        # used for __repr__
        self.class_name = 'ADPosixGroup'
        # posix attributes aren't actually mandatory for posix objects in AD, so we're not guaranteed these
        # exist
        self.gid = UNKNOWN_GROUP_POSIX_GID
        # gidNumber will come back as a singleton list
        if AD_ATTRIBUTE_GID_NUMBER in attributes and attributes.get(AD_ATTRIBUTE_GID_NUMBER):
            self.gid = int(attributes.get(AD_ATTRIBUTE_GID_NUMBER)[0])



from enum import Enum
# limit random passwords to printable characters in case callers retrieve the
# password for use in other applications/manual use.
AD_PASSWORD_CHAR_RANGE = [chr(i) for i in range(33, 127)]


class ADEncryptionType(Enum):
    DES_CBC_CRC = 1
    DES_CBC_MD5 = 2
    RC4_HMAC = 4
    AES128_CTS_HMAC_SHA1_96 = 8
    AES256_CTS_HMAC_SHA1_96 = 16

    @classmethod
    def get_ad_encryption_type_for_value(cls, val):
        return ENCRYPTION_TYPE_VALUE_TO_ENUM.get(val)


# encryption types as stored in AD are just a number that represents a bitstring.
# use this map to convert the number and our strings back and forth
ENCRYPTION_TYPE_VALUE_TO_ENUM = {
    1: ADEncryptionType.DES_CBC_CRC,
    2: ADEncryptionType.DES_CBC_MD5,
    4: ADEncryptionType.RC4_HMAC,
    8: ADEncryptionType.AES128_CTS_HMAC_SHA1_96,
    16: ADEncryptionType.AES256_CTS_HMAC_SHA1_96,
}

# users might want to use string representations of their encryption types. allow for this
# and also allow for the short form of encryption types. MMC and other AD management tools
# often use aes256-sha1 or rc4-hmac even though the actual krb5 encryption types from RFCs
# are longer (aes256-cts-hmac-sha1-96 and arcfour-hmac respectively)
ENCRYPTION_TYPE_STR_TO_ENUM = {
    'des-cbc-crc': ADEncryptionType.DES_CBC_CRC,
    'des-cbc-md5': ADEncryptionType.DES_CBC_MD5,
    'arcfour-hmac': ADEncryptionType.RC4_HMAC,
    'rc4-hmac': ADEncryptionType.RC4_HMAC,
    'aes128-cts-hmac-sha1-96': ADEncryptionType.AES128_CTS_HMAC_SHA1_96,
    'aes128-sha1': ADEncryptionType.AES128_CTS_HMAC_SHA1_96,
    'aes256-cts-hmac-sha1-96': ADEncryptionType.AES256_CTS_HMAC_SHA1_96,
    'aes256-sha1': ADEncryptionType.AES256_CTS_HMAC_SHA1_96,
}

# These encryption types are broken so I assume nobody wants to use them and have not implemented
# support for them. But they're included for completeness so that readers can more easily tell
# what's going on.
UNSUPPORTED_ENC_TYPES = {ADEncryptionType.DES_CBC_CRC, ADEncryptionType.DES_CBC_MD5}



""" Utilities for parsing keytab files into GSS Kerberos Keys. """
import binascii
import os

from ms_active_directory import logging_utils

from ms_active_directory.core.ad_kerberos_keys import GssKerberosKey, RawKerberosKey
# constants for structuring in-memory keytab representations
from ms_active_directory.environment.kerberos.kerberos_constants import (
    KEYTAB_FORMAT_VERSION_FOR_KEYTAB_FORMAT_VERSION,
    KRB5_ENC_TYPE_VALUE_TO_ENC_TYPE_MAP,
)

# constants for parsing keytab data, as a separate import for improved readability
from ms_active_directory.environment.kerberos.kerberos_constants import (
    ENCRYPTION_TYPE_FIELD_SIZE,
    ENTRY_LENGTH_FIELD_SIZE_BYTES,
    FLAGS_FIELD_SIZE_BYTES,
    KEYTAB_STANDARD_LEADING_BYTES_SIZE,
    KEYTAB_FORMAT_SIZE_BYTES,
    KEY_LENGTH_FIELD_SIZE_BYTES,
    NUM_COMPONENTS_FIELD_SIZE_BYTES,
    REALM_LENGTH_FIELD_SIZE_BYTES,
    PRINCIPAL_COMPONENT_DIVIDER,
    PRINCIPAL_COMPONENT_LENGTH_FIELD_SIZE_BYTES,
    PRINCIPAL_TYPE_FIELD_SIZE_BYTES,
    TIMESTAMP_FIELD_SIZE_BYTES,
    VNO8_FIELD_SIZE_BYTES,
    VNO32_FIELD_SIZE_BYTES,
)
from ms_active_directory.exceptions import KeytabEncodingException


logger = logging_utils.get_logger()


def process_keytab_bytes_to_extract_entries(keytab: bytes):
    """ Given a byte string of binary keytab data, extract keytab entries from it and return them as GSSKerberosKeys.
    """
    hex_keytab_data = binascii.hexlify(keytab).decode('utf-8')
    return process_hex_string_keytab_file_to_extract_entries(hex_keytab_data)


def process_keytab_file_to_extract_entries(keytab_file_path: str):
    """ Given a file path for a keytab, extract keytab entries from it and return them as GSSKerberosKeys. """
    if not os.path.isfile(keytab_file_path):
        raise KeytabEncodingException('File {} cannot be found for reading keytabs.'
                                      .format(keytab_file_path))
    keytab_data = open(keytab_file_path, 'rb').read()
    return process_keytab_bytes_to_extract_entries(keytab_data)


def _twos_complement(value: int, bits: int):
    """ This does twos complement so we can convert hex strings to signed integers.
    Why is this not built-in to python int?
    """
    if value & (1 << (bits-1)):
        value -= 1 << bits
    return value


def _read_bytes_as_number(keytab: str, index: int, bytes_to_read: int=1, keytab_format_version: int=1,
                          is_signed_int: bool=False):
    """ Given hex-encoded keytab data, the index we're starting from, the number of
    bytes in the keytab we want to read, and the keytab format version, this function
    will read and interpret the bytes requested starting at the index. Bytes may be
    reordered depending on format version, as versions change between big-endian and
    little-endian encoding.
    The resultant hex is them decoded to an integer and returned.

    Format version 1 means native byte order
    Format version 2 means big-endian byte order
    Format pulled from https://www.h5l.org/manual/HEAD/krb5/krb5_fileformats.html

    A hex number is 4 bits, so our "bytes to read" value gets doubled to determine
    actual offsets in our hex string.
    """
    # since our string is hex, a byte is represented by 2 characters, so our string offset to read is twice
    # the number of bytes
    offset = bytes_to_read * 2
    end_index = index + offset
    if end_index > len(keytab):
        return 0

    hex_string_to_parse = keytab[index:end_index]
    if keytab_format_version == 1:
        converted_from_little_endian = []
        for i in range(0, offset, 2):
            converted_from_little_endian.insert(0, hex_string_to_parse[i:i+2])
        hex_string_to_parse = ''.join(converted_from_little_endian)
    elif keytab_format_version != 2:
        raise KeytabEncodingException('Unrecognized keytab format version {}'.format(keytab_format_version))

    unsigned_value = int(hex_string_to_parse, 16)
    if is_signed_int:
        return _twos_complement(unsigned_value, bytes_to_read * 8)  # 8 bits per byte
    return unsigned_value


def _read_bytes_as_string(keytab: str, index: int, bytes_to_read: int):
    """ Given hex-encoded keytab data, the index we're starting from, the number of
    bytes in the keytab we want to read, and the keytab format version, this function
    will read and interpret the bytes requested starting at the index.
    The resultant hex is them decoded to a UTF-8 string and returned.

    A hex number is 4 bits, so our "bytes to read" value gets doubled to determine
    actual offsets in our hex string.
    """
    offset = bytes_to_read * 2
    end_index = index + offset
    if end_index > len(keytab):
        return '0'  # this is the same as get_bytes_number above. when we can't read, return 0
    return bytearray.fromhex(keytab[index:end_index]).decode('UTF-8')


def _read_bytes_to_number_and_then_move_position(keytab: str, current_keytab_position: int, bytes_to_read: int,
                                                 keytab_format_version: int, is_signed_int: bool=False):
    """ Read some number of bytes from the keytab starting at the given position, move our
    position in the keytab forward, and return the value read as an integer and the new position.

    A hex number is 4 bits, so our "bytes to read" value gets doubled to calculate
    how far to move position in our hex string.
    """
    read_value = _read_bytes_as_number(keytab, index=current_keytab_position, bytes_to_read=bytes_to_read,
                                       keytab_format_version=keytab_format_version, is_signed_int=is_signed_int)
    new_keytab_position = current_keytab_position + 2 * bytes_to_read
    return read_value, new_keytab_position


def _read_bytes_to_string_and_then_move_position(keytab: str, current_keytab_position: int, bytes_to_read: int):
    """ Read some number of bytes from the keytab starting at the given position, move our
    position in the keytab forward, and return the value read and the new position.

    A hex number is 4 bits, so our "bytes to read" value gets doubled to calculate
    how far to move position in our hex string.
    """
    read_value = _read_bytes_as_string(keytab, index=current_keytab_position, bytes_to_read=bytes_to_read)
    new_keytab_position = current_keytab_position + 2 * bytes_to_read
    return read_value, new_keytab_position


def _get_principal_component_length_and_then_read_component(keytab: str, current_keytab_position: int,
                                                            keytab_format_version: int):
    """ Extract the component length value from a keytab and then read the following component.
    Component length is always encoded into the keytab in a fixed size entry, so we can always
    read component lengths the same way and interpret them.

    :returns: a tuple of component value, keytab position after reading
    """
    component_length, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                             PRINCIPAL_COMPONENT_LENGTH_FIELD_SIZE_BYTES,
                                                                                             keytab_format_version)
    component, current_keytab_position = _read_bytes_to_string_and_then_move_position(keytab, current_keytab_position,
                                                                                      component_length)
    return component, current_keytab_position


def process_hex_string_keytab_file_to_extract_entries(keytab):
    """ Given a hex encoded keytab file, extract all of the entries in it and return a list of
    dictionaries describing each entry.
    """
    # keytab metadata - figure out our keytab format version and make sure it's valid
    current_keytab_position = 0
    start_byte = _read_bytes_as_number(keytab, bytes_to_read=KEYTAB_STANDARD_LEADING_BYTES_SIZE,
                                       index=current_keytab_position)
    if start_byte != 5:
        raise KeytabEncodingException('Keytabs must always start with 0x05 as the leading byte, as only Kerberos v5 '
                                      'keytabs are supported. Seen leading byte: {}'.format(start_byte))
    # one byte is 2 hex digits, so move positions equal to 2x our size in bytes
    current_keytab_position += 2*KEYTAB_STANDARD_LEADING_BYTES_SIZE

    # we check our format version by reading bytes in the default v1 format, because that's how
    # backwards compatibility works. the thing that tells you "you're in the new format" is
    # written in the old format, so that everyone can read it.
    keytab_format_version = _read_bytes_as_number(keytab, bytes_to_read=KEYTAB_FORMAT_SIZE_BYTES,
                                                  keytab_format_version=KEYTAB_FORMAT_VERSION_FOR_KEYTAB_FORMAT_VERSION,
                                                  index=current_keytab_position)
    if keytab_format_version != 1 and keytab_format_version != 2:
        raise KeytabEncodingException('Unrecognized and unsupported keytab format version: {}'
                                      .format(keytab_format_version))
    logger.debug('Ingesting keytab with format version %s', keytab_format_version)
    # one byte is 2 hex digits, so move positions equal to 2x our size in bytes
    current_keytab_position += 2*KEYTAB_FORMAT_SIZE_BYTES

    # this is the prefix the encodes our keytab format version. it only appears once for the entire
    # file. so if we want to split up our file into all of its individual entries, we want to prepend
    # the contents of each entry with this standalone prefix so that they can be utilized as file
    # data without extra processing
    standalone_prefix = keytab[:current_keytab_position]
    entries = []
    # int32_t size of entry = 32 bits in entry size = 4 bytes to read
    # entry length is the only signed number in the schema because it can be negative
    entry_length_bytes = _read_bytes_as_number(keytab, index=current_keytab_position,
                                               bytes_to_read=ENTRY_LENGTH_FIELD_SIZE_BYTES,
                                               keytab_format_version=keytab_format_version,
                                               is_signed_int=True)

    # since our string is hex, a byte is represented by 2 characters, so we move our index forwards by twice
    # the number of bytes we read
    current_keytab_position += 2*ENTRY_LENGTH_FIELD_SIZE_BYTES
    # iterate through entries
    slot = 1
    while entry_length_bytes != 0:
        try:
            if entry_length_bytes > 0:
                start_value = current_keytab_position  # start of this entry
                # uint16_t num_components = 16 bits = 2 bytes to read
                num_components = _read_bytes_as_number(keytab, index=current_keytab_position,
                                                       bytes_to_read=NUM_COMPONENTS_FIELD_SIZE_BYTES,
                                                       keytab_format_version=keytab_format_version)
                logger.debug('Reading %s components from keytab entry in slot', num_components, slot)

                # the number of components encoded in a format v1 keytab is 1 greater than it is supposed to be
                if keytab_format_version == 1:
                    num_components -= 1

                if num_components == 0 and entry_length_bytes == 3:
                    raise KeytabEncodingException('Malformed keytab file detected. Slots are not encoded.')

                # since our string is hex, a byte is represented by 2 characters, so we move our index forwards by twice
                # the number of bytes we read
                current_keytab_position += 2*NUM_COMPONENTS_FIELD_SIZE_BYTES

                # counted octet string realm (prefixed with 16bit length, no null terminator)
                realm_length, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                                     REALM_LENGTH_FIELD_SIZE_BYTES,
                                                                                                     keytab_format_version)
                if realm_length == 0:
                    raise KeytabEncodingException('Malformed keytab file detected. A realm length of 0 is encoded to a '
                                                  'keytab.')
                realm, current_keytab_position = _read_bytes_to_string_and_then_move_position(keytab, current_keytab_position,
                                                                                              realm_length)

                principal_components = []
                for i in range(num_components):
                    piece, current_keytab_position = _get_principal_component_length_and_then_read_component(keytab, current_keytab_position,
                                                                                                             keytab_format_version)
                    principal_components.append(piece)
                # principal components are separated by forwarded slashes (not encoded)
                # generally, we're likely to only have 1 or 2 components; we have 1 for user keytabs, because the
                # principal is just the username used to authenticate, and we have 2 for most server keytabs, because
                # the principal is a combination of the "service" (e.g. ssh, ldap, nfs) and the FQDN that should be used
                # to address the server during authentication.
                principal = PRINCIPAL_COMPONENT_DIVIDER.join(principal_components)

                # uint32_t name_type = 32 bits in name_type = 4 bytes to read
                # name type is not included in format version 1 keytabs
                if keytab_format_version != 1:
                    name_type, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                                      PRINCIPAL_TYPE_FIELD_SIZE_BYTES,
                                                                                                      keytab_format_version)
                else:
                    name_type = 1

                # uint32_t timestamp (time key was established) = 32 bits in time = 4 bytes to read
                timestamp, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                                  TIMESTAMP_FIELD_SIZE_BYTES,
                                                                                                  keytab_format_version)

                # uint8_t vno8 = 8 bits in kvno = 1 byte to read
                vno8, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                             VNO8_FIELD_SIZE_BYTES,
                                                                                             keytab_format_version)
                vno = vno8

                # keyblock structure: 16-bit value for encryption type and then counted_octet for key
                encryption_type, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                                        ENCRYPTION_TYPE_FIELD_SIZE,
                                                                                                        keytab_format_version)

                key_length, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                                   KEY_LENGTH_FIELD_SIZE_BYTES,
                                                                                                   keytab_format_version)

                # we leave our key hex encoded so we don't use read_number_and_move_position
                # since our string is hex, a byte is represented by 2 characters, so we our end index for reading our
                # key is twice its length in bytes
                hex_encoded_key = keytab[current_keytab_position:current_keytab_position+(key_length*2)]
                current_keytab_position += key_length * 2

                # uint32_t vno if >=4 bytes left in entry_length_bytes
                current_entry_length_bytes = (current_keytab_position - start_value)//2
                if entry_length_bytes - current_entry_length_bytes >= VNO32_FIELD_SIZE_BYTES:
                    vno32, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                                  VNO32_FIELD_SIZE_BYTES,
                                                                                                  keytab_format_version)
                    # We will always pick the 32-bit vno's value if it is non-zero. Due to padding all entries in a
                    # keytab file to be the same length, the 32-bit vno field can be 0 if it's not actually populated
                    # at all (e.g. if it was generated on an older machine that didn't encode it). That's why we skip
                    # the value if it's 0
                    # https://web.mit.edu/kerberos/www/krb5-latest/doc/formats/keytab_file_format.html
                    if vno32 != 0:
                        vno = vno32

                # uint32_t flags if >=4 bytes left in entry_length_bytes
                flags = None
                current_entry_length_bytes = (current_keytab_position - start_value)//2
                if entry_length_bytes - current_entry_length_bytes >= FLAGS_FIELD_SIZE_BYTES:
                    flags, current_keytab_position = _read_bytes_to_number_and_then_move_position(
                        keytab, current_keytab_position,
                        FLAGS_FIELD_SIZE_BYTES, keytab_format_version)

                # a keytab might not take up the full value of its entry because a tool may have overwritten an old
                # entry with a newer, smaller one. this leaves space behind. (this is also what can cause whitespace
                # to be introduced into principal names)
                # in this case, there's zero padding to the end of the entry that we can skip
                # see: https://web.mit.edu/kerberos/www/krb5-latest/doc/formats/keytab_file_format.html
                # we do this before computing our single entry value so that the entry length encoded is still accurate
                # after parsing.
                # if the dictionary details are written to data, we'll recompute the minimal length
                # and store that instead. but for the purpose of parsing, we want to return the data exactly as it
                # exists in the keytab file without alteration
                current_entry_length_bytes = (current_keytab_position - start_value)//2
                if current_entry_length_bytes < entry_length_bytes:
                    bytes_to_move = entry_length_bytes - current_entry_length_bytes
                    # 1 byte = 2 hex characters
                    current_keytab_position += bytes_to_move*2

                # subtract 8 from our start value to reintroduce the specification of how
                # long the keytab is, which we jump over before starting the loop.
                # this will become the bytes that we'd write to a file if we wanted to extract this
                # entry in the keytab into its own file, so we need to include the keytab length as
                # well as the prefix indicating which version of keytab format we're using in the
                # value
                single_entry_value = standalone_prefix + keytab[start_value-8:current_keytab_position]

                # strip any whitespace from the front of the principal that may have gotten
                # there when moving keytabs to canonical form, but which doesn't matter
                # from a functional perspective
                principal = principal.lstrip()
                string_enc_type = KRB5_ENC_TYPE_VALUE_TO_ENC_TYPE_MAP[encryption_type]
                raw_key = RawKerberosKey(string_enc_type, binascii.unhexlify(hex_encoded_key))
                keytab_entry = GssKerberosKey(principal, realm, raw_key, vno,
                                              flags, timestamp, name_type, keytab_format_version)
                entries.append(keytab_entry)
            else:
                # 0 length or negative length keytabs indicate deleted entries that were not erased from the file
                # so we skip them
                logger.debug('Skipping %s length keytab due to indication of a deleted entry', abs(entry_length_bytes))
                current_keytab_position += abs(entry_length_bytes) * 2
        finally:
            entry_length_bytes, current_keytab_position = _read_bytes_to_number_and_then_move_position(keytab, current_keytab_position,
                                                                                                       ENTRY_LENGTH_FIELD_SIZE_BYTES,
                                                                                                       keytab_format_version,
                                                                                                       is_signed_int=True)
            slot += 1
    logger.debug('Extracted %s kerberos keys from keytab', len(entries))
    return entries
